from utils.utils import TrainableWeightsReshaper
import tensorflow as tf
import numpy as np


class ActorCriticAgent:
    """
    An actor critic agent which learns from the latest step in a history (one-step update)
    """

    def __init__(self, actor, critic, gamma, agent_uid, name=""):
        """
        :param actor: An actor which calculates policy and predict actions
        :param critic: A critic which estimates the agent's performance
        :param gamma: A discount factor
        :param agent_uid: A unique identification of the agent
        :param name: The name of the agent
        """
        self.actor = actor
        self.critic = critic
        self.gamma = gamma
        self.name = name
        self.agent_uid = agent_uid

    def next_action(self, state, available_actions):
        """
        Predict the next action.
        :param state: The state to predict the action from
        :param available_actions: The available actions
        :return: An action
        """
        return self.actor.next_action(state, available_actions)

    def learn(self, history):
        step_discount = self.calculate_step_discount(history.get_current_step(self.agent_uid))
        value_target = self.calculate_value_targets(history.get_current_reward(self.agent_uid),
                                                    history.get_next_state(self.agent_uid),
                                                    history.is_currently_done(self.agent_uid))
        td_error = self.calculate_td_error(history.get_current_state(self.agent_uid), value_target)
        self.update_critic(history.get_current_state(self.agent_uid), value_target)
        self.update_actor(history.get_current_state(self.agent_uid),
                          history.get_currently_available_actions(self.agent_uid),
                          history.get_current_action(self.agent_uid),
                          td_error, step_discount)
        return step_discount, value_target, td_error

    def update_critic(self, states, value_targets):
        self.critic.update_value_function(states, value_targets)

    def update_actor(self, states, available_actions, actions, td_error, step_discount):
        self.actor.update_policy_function_rl(states, available_actions, actions, td_error, step_discount)

    def calculate_step_discount(self, steps):
        return self.gamma ** (steps - 1)

    def calculate_value_targets(self, rewards, next_states, is_done):
        return rewards + self.gamma * self.critic.calculate_value(next_states) * (1 - is_done)

    def calculate_td_error(self, states, value_target):
        return value_target - self.critic.calculate_value(states)

    def get_actor_weights(self):
        return self.actor.policy.get_weights()

    def set_actor_weights(self, weights):
        self.actor.policy.set_weights(weights)


class BatchActorCriticAgent(ActorCriticAgent):
    """
    An actor-critic agent which learns from a history.
    """

    def learn(self, history):
        step_discounts = self.calculate_step_discount(history.get_steps(self.agent_uid))
        value_targets = self.calculate_value_targets(history.get_rewards(self.agent_uid),
                                                     history.get_next_states(self.agent_uid),
                                                     history.is_done(self.agent_uid))
        td_errors = self.calculate_td_error(history.get_states(self.agent_uid), value_targets)
        self.update_critic(history.get_states(self.agent_uid), value_targets)
        self.update_actor(history.get_states(self.agent_uid),
                          history.get_available_actions(self.agent_uid),
                          history.get_actions(self.agent_uid),
                          td_errors, step_discounts)
        return step_discounts, value_targets, td_errors


class ActorCriticLOLAAgent(ActorCriticAgent):
    """
    An actor critic agent similar to an ActorCriticAgent, except that it also learns by applying
    the LOLA-correction
    """

    def __init__(self, actor, critic, gamma, opponents_lr, agent_uid, name=""):
        super().__init__(actor, critic, gamma, agent_uid, name=name)
        self.opponent_actors = None
        self.opponents_learning_rate = opponents_lr
        self.reshaper = TrainableWeightsReshaper(self.actor.get_trainable_weights())
        self.opponent_modelling = True
        self.agent_uid = agent_uid

    def add_opponent_actors(self, opponent_actors, opponent_modelling=True):
        self.opponent_modelling = opponent_modelling
        self.opponent_actors = opponent_actors

    def learn(self, history):
        step_discount, value_target, td_error = super().learn(history)
        if self.opponent_modelling:
            self.update_opponent_actors(history)
        self.calculate_lola_corrections_and_update_actor(history, td_error, step_discount)

    def update_opponent_actors(self, history):
        for opponent in self.opponent_actors.keys():
            self.update_opponent_actor(self.opponent_actors[opponent],
                                       history.get_current_state(opponent),
                                       history.get_currently_available_actions(opponent),
                                       history.get_current_action(opponent))

    def calculate_lola_corrections_and_update_actor(self, history, td_error, step_discount):
        lola_gradient = self.calculate_lola_gradient(history, td_error, step_discount) * self.opponents_learning_rate
        lola_gradient = self.reshaper.restore(lola_gradient)
        self.actor.apply_gradients(lola_gradient)

    def update_opponent_actor(self, opponent_actor, states, available_actions, actions):
        opponent_actor.update_policy_function_sl(states, available_actions, actions)

    def calculate_lola_gradient(self, history, td_error, step_discount):
        opponent_corrections = [self.calculate_lola_gradient_single_opponent(history, opponent_id, td_error,
                                                                             step_discount) for opponent_id in
                                self.opponent_actors.keys()]
        return tf.squeeze(tf.reduce_sum(opponent_corrections, axis=0))

    def calculate_lola_gradient_single_opponent(self, history, opponent_uid, td_error, step_discount):
        opponent_actor = self.opponent_actors[opponent_uid]
        opponent_reward = history.get_current_reward(opponent_uid)
        negative_policy_gradient_opponent_params = self.reshaper.flatten(opponent_actor.calculate_policy_gradient(
            history.get_current_state(self.agent_uid),
            history.get_currently_available_actions(opponent_uid),
            history.get_current_action(opponent_uid),
            td_error,
            step_discount,
            training=True))
        self_log_pi_gradient = self.reshaper.flatten(
            self.actor.calculate_log_pi_gradient(history.get_states(self.agent_uid),
                                                 history.get_available_actions(self.agent_uid),
                                                 history.get_actions(self.agent_uid), training=True))

        opponent_log_pi_gradient = self.reshaper.flatten(
            opponent_actor.calculate_log_pi_gradient(history.get_states(opponent_uid),
                                                     history.get_available_actions(opponent_uid),
                                                     history.get_actions(opponent_uid), training=True))
        lola_correction = tf.expand_dims(negative_policy_gradient_opponent_params, axis=0) @ (
                step_discount * opponent_reward * tf.transpose(tf.expand_dims(self_log_pi_gradient, axis=0))) @ \
                          tf.expand_dims(opponent_log_pi_gradient, axis=0)

        return lola_correction


class BatchActorCriticLOLAAgent(BatchActorCriticAgent, ActorCriticLOLAAgent):
    """
        An actor critic agent similar to an BatchActorCriticAgent, except that it also learns by applying
        the LOLA-correction
    """

    def learn(self, history):
        step_discount, value_target, td_error = super().learn(history)
        if self.opponent_modelling:
            self.update_opponent_actors(history)
        self.calculate_lola_corrections_and_update_actor(history, td_error, step_discount)

    def update_opponent_actors(self, history):
        for opponent in self.opponent_actors.keys():
            self.update_opponent_actor(self.opponent_actors[opponent],
                                       history.get_states(opponent),
                                       history.get_available_actions(opponent),
                                       history.get_actions(opponent))

    def calculate_lola_gradient_single_opponent(self, history, opponent_uid, td_error, step_discount):
        opponent_actor = self.opponent_actors[opponent_uid]
        opponent_reward = history.get_rewards(opponent_uid)
        negative_policy_gradient_opponent_params = self.reshaper.flatten(opponent_actor.calculate_policy_gradient(
            history.get_states(opponent_uid),
            history.get_available_actions(opponent_uid),
            history.get_actions(opponent_uid),
            td_error,
            step_discount))
        # adapted from https://github.com/alshedivat/lola/blob/master/lola/corrections.py
        with tf.GradientTape() as tape1:
            with tf.GradientTape() as tape2:
                self_log_pi = self.actor.calculate_log_pi(history.get_states(self.agent_uid),
                                                          history.get_available_actions(self.agent_uid),
                                                          history.get_actions(self.agent_uid), training=True)
                opponent_log_pi = opponent_actor.calculate_log_pi(history.get_states(opponent_uid),
                                                                  history.get_available_actions(opponent_uid),
                                                                  history.get_actions(opponent_uid),
                                                                  training=True)
                self_log_pi_cumsum = tf.cumsum(self_log_pi)
                opponent_log_pi_cumsum = tf.cumsum(opponent_log_pi)
                v1 = tf.reduce_sum(step_discount * opponent_reward * self_log_pi_cumsum * opponent_log_pi_cumsum)
            v1_self_gradient = self.reshaper.flatten(tape2.gradient(v1, self.actor.get_trainable_weights()))
            multiply = tf.expand_dims(negative_policy_gradient_opponent_params, axis=0) @ tf.reshape(
                v1_self_gradient, [-1, 1])
        lola_correction = self.reshaper.flatten(tape1.gradient(multiply, opponent_actor.get_trainable_weights()))
        return lola_correction


class DefaultCritic:

    def __init__(self, value_function):
        self.value_function = value_function

    def calculate_value(self, state, training=None):
        return self.value_function(state, training=training)

    def update_value_function(self, state, value_target):
        gradients = self.calculate_value_function_gradient(state, value_target, training=True)
        self.apply_gradients(gradients)

    def calculate_value_function_gradient(self, state, value_target, training=None):
        with tf.GradientTape() as tape:
            predicted_value = self.calculate_value(state, training=training)
            loss = self.calculate_loss(value_target, predicted_value)
        return tape.gradient(loss, self.get_trainable_weights())

    def calculate_loss(self, value_target, predicted_value):
        return self.value_function.loss(value_target, predicted_value)

    def get_trainable_weights(self):
        return self.value_function.trainable_weights

    def apply_gradients(self, gradients):
        self.value_function.optimizer.apply_gradients(zip(gradients, self.value_function.trainable_weights))


class DefaultActor:

    def __init__(self, policy):
        self.policy = policy

    def next_action(self, state, available_actions):
        action_probs = self.policy(tf.expand_dims(state, axis=0))
        return np.random.choice(len(available_actions), p=tf.squeeze(action_probs).numpy())

    def update_policy_function_sl(self, state, available_actions, action):
        gradients = self.calculate_sl_gradient(state, available_actions, action, training=True)
        self.apply_gradients(gradients)

    def update_policy_function_rl(self, state, available_actions, action, td_error, step_discount):
        gradients = self.calculate_policy_gradient(state, available_actions, action, td_error, step_discount,
                                                   training=True)
        self.apply_gradients(gradients)

    def calculate_sl_gradient(self, state, available_actions, action, training=None):
        with tf.GradientTape() as tape:
            loss = self.calculate_sl_loss(state, available_actions, action, training=training)
        return tape.gradient(loss, self.get_trainable_weights())

    def calculate_sl_loss(self, state, available_actions, action, training=None):
        return self.policy.loss(action, self.policy(state, training=training))

    def calculate_policy_gradient(self, state, available_actions, action, td_error, step_discount, training=None):
        with tf.GradientTape() as tape:
            policy_loss = - td_error * step_discount * self.calculate_log_pi(state, available_actions, action,
                                                                             training=training)
        return tape.gradient(policy_loss, self.get_trainable_weights())

    def calculate_log_pi_gradient(self, state, available_actions, action, training=None):
        with tf.GradientTape() as tape:
            policy_loss = self.calculate_log_pi(state, available_actions, action, training=training)
        return tape.gradient(policy_loss, self.get_trainable_weights())

    def get_trainable_weights(self):
        return self.policy.trainable_weights

    def apply_gradients(self, gradients):
        self.policy.optimizer.apply_gradients(zip(gradients, self.policy.trainable_weights))

    def calculate_log_pi(self, states, available_actions, actions, training=None):
        probs = self.policy(states)
        chosen_action_probs = tf.gather(probs, actions, batch_dims=1)
        return tf.math.log(chosen_action_probs)


class DiplomacyCritic(DefaultCritic):

    def __init__(self, value_function, power_index):
        """
        :param value_function:
        :param power_index: The index of the value function's output which represents the state value
        of the power the critic is calculating state values for.
        """
        super().__init__(value_function)
        self.power_index = power_index

    def set_power_index(self, power_index):
        self.power_index = power_index

    def calculate_loss(self, value_target, predicted_value):
        return self.value_function.value_function_loss(value_target, predicted_value)

    def calculate_value(self, states, training=None):
        return self.value_function(states, training=training)[:,
               self.power_index:self.power_index + 1]


class DiplomacyActor(DefaultActor):

    def __init__(self, policy):
        super().__init__(policy)

    def next_action(self, state, available_actions):
        chosen_orders, chosen_orders_probs = self.policy([state, available_actions])
        return chosen_orders[0]

    def calculate_log_pi(self, states, available_actions, actions, training=None):
        chosen_orders, chosen_orders_probs = self.policy([states, available_actions], actions=actions,
                                                         training=training)
        prob = tf.reduce_prod(chosen_orders_probs, axis=1)
        log_prob = tf.math.log(prob + 1e-37)  # 1e-37 is added avoid -inf values
        return log_prob


class RandomDiplomacyActor(DefaultActor):
    """
    A Diplomacy actor which chooses actions randomly
    """

    def __init__(self, default_action_index):
        super().__init__(None)
        self.default_action_index = default_action_index

    def next_action(self, state, available_actions):
        available_actions_encoding, power = available_actions

        assert len(available_actions_encoding.shape) == 4
        available_actions_encoding = available_actions_encoding[..., 1][0]
        chosen_actions = []
        for province_actions in available_actions_encoding:
            available_actions_in_province = self.remove_default_actions(province_actions)
            if available_actions_in_province.size != 0:
                chosen_actions.append(np.random.choice(available_actions_in_province, size=(1,)))
        return np.array(chosen_actions)

    def remove_default_actions(self, available_actions):
        return np.delete(available_actions, np.where(available_actions == self.default_action_index))


class AgentFactory:
    AGENT_UID = 0

    @staticmethod
    def create_agent(agent_type, actor_type="DEFAULT", critic_type="DEFAULT", **agent_config):
        """
        :param agent_type: Type of agent. Available: A2C, LOLA, BATCH_A2C, BATCH_LOLA
        :param actor_type: Type of actor. Available: DIPLOMACY, RANDOM_DIPLOMACY, DEFAULT
        :param critic_type: Type of critic. Available: DIPLOMACY, DEFAULT
        :param agent_config: Parameter values which is needed to create the agent
        :return: An agent
        """
        if actor_type == "DIPLOMACY":
            actor = DiplomacyActor(agent_config["pnet"])
        elif actor_type == "RANDOM_DIPLOMACY":
            actor = RandomDiplomacyActor(agent_config["default_action_index"])
        else:
            actor = DefaultActor(agent_config["pnet"])
        if critic_type == "DIPLOMACY":
            critic = DiplomacyCritic(agent_config["vnet"], AgentFactory.AGENT_UID)
        else:
            critic = DefaultCritic(agent_config["vnet"])
        if agent_type == "BATCH_LOLA":
            agent = BatchActorCriticLOLAAgent(actor=actor,
                                              critic=critic,
                                              gamma=agent_config["gamma"],
                                              opponents_lr=agent_config["opponent_learning_rate"],
                                              agent_uid=AgentFactory.AGENT_UID,
                                              name=agent_config["agent_name"])

        elif agent_type == "BATCH_A2C":
            agent = BatchActorCriticAgent(actor=actor,
                                          critic=critic,
                                          gamma=agent_config["gamma"],
                                          agent_uid=AgentFactory.AGENT_UID,
                                          name=agent_config["agent_name"])

        elif agent_type == "LOLA":
            agent = ActorCriticLOLAAgent(actor=actor,
                                         critic=critic,
                                         gamma=agent_config["gamma"],
                                         opponents_lr=agent_config["opponent_learning_rate"],
                                         agent_uid=AgentFactory.AGENT_UID,
                                         name=agent_config["agent_name"])

        elif agent_type == "A2C":
            agent = ActorCriticAgent(actor=actor,
                                     critic=critic,
                                     gamma=agent_config["gamma"],
                                     agent_uid=AgentFactory.AGENT_UID,
                                     name=agent_config["agent_name"])
        else:
            raise ValueError("The agent cannot be provided")
        AgentFactory.AGENT_UID += 1
        return agent

    @staticmethod
    def create_list_of_agents(agent_type, pnet, vnet, agent_configs):
        agents = [AgentFactory.create_agent(agent_type=agent_type,
                                            pnet=pnet,
                                            vnet=vnet,
                                            **single_agent_config) for single_agent_config in agent_configs]
        add_opponent_actors(agents)
        return agents


def add_opponent_actors(lola_agents):
    """
    Add opponent actors to each lola agent in the lola_agents list.
    The opponent actors are the other lola agents' actors.
    :param lola_agents: List of lola agents
    :return:
    """
    for agent1 in lola_agents:
        if hasattr(agent1, "add_opponent_actors"):
            opponent_actors = {}
            for agent2 in lola_agents:
                if agent1 != agent2:
                    opponent_actors[agent2.agent_uid] = agent2.actor
            agent1.add_opponent_actors(opponent_actors, False)
