import unittest
from unittest.mock import Mock
import tensorflow as tf
import numpy as np
from agents.actor_critic_agents import DefaultActor, ActorCriticAgent, ActorCriticLOLAAgent, BatchActorCriticAgent, \
    BatchActorCriticLOLAAgent, DiplomacyActor, AgentFactory, DefaultCritic, DiplomacyCritic, RandomDiplomacyActor, \
    add_opponent_actors


class TestActorCriticAgent(unittest.TestCase):

    def test(self):
        actor = Mock()
        actor.next_action = Mock(return_value=2)
        a2c_agent = ActorCriticAgent(actor, None, 0.99, 2)
        self.assertEqual(a2c_agent.next_action(2, 2), 2)


class TestActorCriticLOLAAgent(unittest.TestCase):

    def setUp(self) -> None:
        self.actor1 = DefaultActor(SuperSimplePolicy(tf.Variable(np.array([[3], [5]]), dtype=tf.float32)))
        self.actor2 = DefaultActor(SuperSimplePolicy(tf.Variable(np.array([[1], [2]]), dtype=tf.float32)))

        self.player1 = ActorCriticLOLAAgent(self.actor1, None, 0.96, 0.05, 0)
        self.player2 = ActorCriticLOLAAgent(self.actor2, None, 0.96, 0.05, 1)

        self.player1.add_opponent_actors({self.player2.agent_uid: self.actor2})

    def test_calculate_lola_gradient_one_step(self):
        history = Mock()
        history.get_current_state = Mock(return_value=np.array([[2., 3.]]))
        history.get_currently_available_actions = Mock(return_value=[])
        history.get_current_action = Mock(return_value=np.array([[0]]))
        history.get_current_reward = Mock(return_value=np.array([[2.]]))
        history.get_states = Mock(return_value=np.array([[2., 3.]]))
        history.get_available_actions = Mock(return_value=[])
        history.get_actions = Mock(return_value=np.array([[0]]))

        lola_correction = self.player1.calculate_lola_gradient(history, np.array([[3.]]), 1.0)

        np.testing.assert_almost_equal(lola_correction, np.array([-0.11607142857142858, -0.17410714285714285]))

    def test_calculate_lola_gradient_two_steps(self):
        history = Mock()
        history.get_current_state = Mock(return_value=np.array([[1., 2.]]))
        history.get_currently_available_actions = Mock(return_value=[])
        history.get_current_action = Mock(return_value=np.array([[0]]))
        history.get_current_reward = Mock(return_value=np.array([[3.]]))
        history.get_states = Mock(return_value=np.array([[2., 3.],
                                                         [1., 2.]]))
        history.get_available_actions = Mock(return_value=[])
        history.get_actions = Mock(return_value=np.array([[0], [0]]))

        second_order = 0.96 * 3 * np.array(
            [[(2 / 21 + 1 / 13) * (2 / 8 + 1 / 5), (2 / 21 + 1 / 13) * (3 / 8 + 2 / 5)],
             [(3 / 21 + 2 / 13) * (2 / 8 + 1 / 5), (3 / 21 + 2 / 13) * (3 / 8 + 2 / 5)]])

        first_order_opponent = np.array([[-0.96 * 1 / 5, -0.96 * 2 / 5]])
        expected_gradient = np.squeeze(first_order_opponent @ second_order)

        lola_correction = self.player1.calculate_lola_gradient(history, np.array([[1.]]), 0.96)
        np.testing.assert_almost_equal(lola_correction, expected_gradient, decimal=6)


class TestBatchActorCriticLOLAAgent(unittest.TestCase):

    def setUp(self) -> None:
        self.actor1 = DefaultActor(SuperSimplePolicy(tf.Variable(np.array([[3], [5]]), dtype=tf.float32)))
        self.actor2 = DefaultActor(SuperSimplePolicy(tf.Variable(np.array([[1], [2]]), dtype=tf.float32)))

        self.player1 = BatchActorCriticLOLAAgent(self.actor1, None, 0.96, 0.05, 0)
        self.player2 = BatchActorCriticLOLAAgent(self.actor2, None, 0.96, 0.05, 1)

        self.player1.add_opponent_actors({self.player2.agent_uid: self.actor2})

    def test_calculate_lola_gradient_one_step(self):
        td_error = tf.Variable(3., dtype=tf.float32)
        step_discount = tf.Variable(1, dtype=tf.float32)

        history = Mock()
        history.get_rewards = Mock(return_value=np.array([[2]]))
        history.get_states = Mock(return_value=np.array([[2., 3.]]))
        history.get_available_actions = Mock(return_value=[])
        history.get_actions = Mock(return_value=np.array([[0]]))

        lola_correction = self.player1.calculate_lola_gradient(history,
                                                               td_error,
                                                               step_discount)

        # expected is calculated by hand
        np.testing.assert_almost_equal(lola_correction, np.array([-0.11607142857142858, -0.17410714285714285]))

    def test_calculate_lola_gradient_two_steps(self):
        td_errors = tf.Variable(np.array([[3.], [1.]]), dtype=tf.float32)
        step_discount = tf.Variable([[1.0], [0.96]], dtype=tf.float32)

        history = Mock()
        history.get_rewards = Mock(return_value=np.array([[2.], [3.]]))
        history.get_states = Mock(return_value=np.array([[2., 3.],
                                                         [1., 2.]]))
        history.get_available_actions = Mock(return_value=[])
        history.get_actions = Mock(return_value=np.array([[0], [0]]))

        second_order = 0.96 * 3 * np.array(
            [[(2 / 21 + 1 / 13) * (2 / 8 + 1 / 5), (2 / 21 + 1 / 13) * (3 / 8 + 2 / 5)],
             [(3 / 21 + 2 / 13) * (2 / 8 + 1 / 5), (3 / 21 + 2 / 13) * (3 / 8 + 2 / 5)]]) + \
                       (2. * np.array([[2 / 21], [3 / 21]]) @ np.array([[2 / 8, 3 / 8]]))

        first_order_opponent = -np.array([[0.96 * 1 / 5 + 3 * 2 / 8, 0.96 * 2 / 5 + 3 * 3 / 8]])
        expected_gradient = np.squeeze(first_order_opponent @ second_order)
        lola_correction = self.player1.calculate_lola_gradient(history,
                                                               td_errors,
                                                               step_discount)
        np.testing.assert_almost_equal(lola_correction, expected_gradient, decimal=6)


class TestDiplomacyCritic(unittest.TestCase):

    def test_calculate_value(self):
        value_function = Mock(return_value=np.array([[0, 1, 2, 3, 4]]))
        for i in range(5):
            critic = DiplomacyCritic(value_function, power_index=i)
            np.testing.assert_equal(critic.calculate_value(2), [[i]])


class TestDefaultActor(unittest.TestCase):

    def setUp(self):
        self.actor = DefaultActor(SuperSimplePolicy(tf.Variable(np.array([[3], [5]]), dtype=tf.float32)))

    def test_calculate_log_pi_single_state(self):
        states = np.array([[4, 5]])
        actions = np.array([[0]])
        log_pi = self.actor.calculate_log_pi(states, [], actions)
        np.testing.assert_almost_equal(log_pi, np.array([[np.log(3 * 4 + 5 * 5)]]))

    def test_calculate_log_pi_multiple_states(self):
        states = np.array([[4, 5], [2, 3], [10, 3]])
        actions = np.array([[0], [0], [0]])
        log_pi = self.actor.calculate_log_pi(states, [], actions)
        np.testing.assert_almost_equal(log_pi, np.log(np.array([[3 * 4 + 5 * 5], [3 * 2 + 5 * 3], [3 * 10 + 5 * 3]])))

    def test_calculate_log_pi_gradient(self):
        state = np.array([[1., 2.],
                          [2., 3.]])
        gradient = self.actor.calculate_log_pi_gradient(state, [], action=np.array([[0], [0]]))
        np.testing.assert_almost_equal(np.array([[2 / 21 + 1 / 13],
                                                 [3 / 21 + 2 / 13]]), gradient[0])

    def test_calculate_log_pi_gradient_correct_action_is_chosen(self):
        policy = Mock(return_value=np.array([[np.e, np.e ** 2], [np.e, np.e ** 2]]))
        da = DefaultActor(policy)
        log_pi = da.calculate_log_pi(None, None, np.array([[0], [1]]))
        np.testing.assert_almost_equal(np.array([[1], [2]]), log_pi)


class TestRandomDiplomacyActor(unittest.TestCase):

    def test_next_action(self):
        available_actions = (np.array([[[[0, 20], [0, 2], [0, 3]], [[1, 30], [1, 30], [1, 15]],
                                      [[2, 30], [2, 30], [2, 30]]]]), 1)
        random_actor = RandomDiplomacyActor(30)
        chosen_actions = random_actor.next_action(None, available_actions)
        self.assertEqual((2, 1), chosen_actions.shape)
        self.assertTrue(chosen_actions[0] in [20, 2, 3])
        self.assertEqual(15, chosen_actions[1])


class TestDiplomacyActor(unittest.TestCase):

    def test_next_action(self):
        chosen_orders = np.array([[[0], [3], [5]]])
        chosen_orders_probs = np.array([[[0.25], [0.25], [0.5]]])
        policy = Mock(return_value=(chosen_orders, chosen_orders_probs))
        actor = DiplomacyActor(policy)
        np.testing.assert_equal([[0], [3], [5]], actor.next_action(None, None))

    def test_calculate_log_pi(self):
        chosen_orders = np.array([[[0], [3], [5]], [[2], [3], [4]]])
        chosen_orders_probs = np.array([[[0.25], [0.25], [0.5]],
                                        [[0.4], [0.4], [0.4]]])
        policy = Mock(return_value=(chosen_orders, chosen_orders_probs))
        actor = DiplomacyActor(policy)
        expected = np.log([[0.25 ** 2 * 0.5], [0.4 ** 3]])
        np.testing.assert_almost_equal(expected, actor.calculate_log_pi(None, None, None))

    def test_calculate_log_pi_zero_prob(self):
        chosen_orders = np.array([[[0], [3], [5]]])
        chosen_orders_probs = np.array([[[0.0], [0.0], [0.0]]])
        policy = Mock(return_value=(chosen_orders, chosen_orders_probs))
        actor = DiplomacyActor(policy)
        log_pi = actor.calculate_log_pi(None, None, None)
        self.assertTrue(log_pi[0, 0] < -85)


class TestAgentFactory(unittest.TestCase):

    def setUp(self):
        self.pnet = Mock()
        self.pnet.trainable_weights = [np.array([[2., 3.]])]
        self.vnet = Mock()
        self.gamma = 0.99
        self.default_action_index = 5
        self.agent_config = {
            "pnet": self.pnet,
            "vnet": self.vnet,
            "gamma": self.gamma,
            "default_action_index": self.default_action_index,
            "agent_name": "",
            "opponent_learning_rate": 0.05
        }

    def test_a2c_default_actor_and_critic(self):
        agent = AgentFactory.create_agent("A2C", **self.agent_config)
        self.assertEqual(type(agent), ActorCriticAgent)
        self.assertEqual(type(agent.actor), DefaultActor)
        self.assertEqual(type(agent.critic), DefaultCritic)

    def test_a2c_diplomacy_actor_and_critic(self):
        agent = AgentFactory.create_agent("A2C", critic_type="DIPLOMACY",
                                          actor_type="DIPLOMACY", **self.agent_config)
        self.assertEqual(type(agent), ActorCriticAgent)
        self.assertEqual(type(agent.actor), DiplomacyActor)
        self.assertEqual(type(agent.critic), DiplomacyCritic)

    def test_a2c_random_diplomacy_actor(self):
        agent = AgentFactory.create_agent("A2C", actor_type="RANDOM_DIPLOMACY", **self.agent_config)
        self.assertEqual(type(agent), ActorCriticAgent)
        self.assertEqual(type(agent.actor), RandomDiplomacyActor)
        self.assertEqual(type(agent.critic), DefaultCritic)

    def test_batch_a2c_default_actor_and_critic(self):
        agent = AgentFactory.create_agent("BATCH_A2C", **self.agent_config)
        self.assertEqual(type(agent), BatchActorCriticAgent)
        self.assertEqual(type(agent.actor), DefaultActor)
        self.assertEqual(type(agent.critic), DefaultCritic)

    def test_batch_a2c_diplomacy_actor_and_critic(self):
        agent = AgentFactory.create_agent("BATCH_A2C", actor_type="DIPLOMACY",
                                          critic_type="DIPLOMACY", **self.agent_config)
        self.assertEqual(type(agent), BatchActorCriticAgent)
        self.assertEqual(type(agent.actor), DiplomacyActor)
        self.assertEqual(type(agent.critic), DiplomacyCritic)

    def test_batch_lola_default_actor_and_critic(self):
        agent = AgentFactory.create_agent("BATCH_LOLA", **self.agent_config)
        self.assertEqual(type(agent), BatchActorCriticLOLAAgent)
        self.assertEqual(type(agent.actor), DefaultActor)
        self.assertEqual(type(agent.critic), DefaultCritic)

    def test_batch_lola_diplomacy_actor_and_critic(self):
        agent = AgentFactory.create_agent("BATCH_LOLA", actor_type="DIPLOMACY",
                                          critic_type="DIPLOMACY", **self.agent_config)
        self.assertEqual(type(agent), BatchActorCriticLOLAAgent)
        self.assertEqual(type(agent.actor), DiplomacyActor)
        self.assertEqual(type(agent.critic), DiplomacyCritic)

    def test_lola_default_actor_and_critic(self):
        agent = AgentFactory.create_agent("LOLA", **self.agent_config)
        self.assertEqual(type(agent), ActorCriticLOLAAgent)
        self.assertEqual(type(agent.actor), DefaultActor)
        self.assertEqual(type(agent.critic), DefaultCritic)

    def test_lola_diplomacy_actor_and_critic(self):
        agent = AgentFactory.create_agent("LOLA", critic_type="DIPLOMACY",
                                          actor_type="DIPLOMACY", **self.agent_config)
        self.assertEqual(type(agent), ActorCriticLOLAAgent)
        self.assertEqual(type(agent.actor), DiplomacyActor)
        self.assertEqual(type(agent.critic), DiplomacyCritic)


class TestAddOpponentActors(unittest.TestCase):

    def test_add_opponent_actors(self):
        def create_lola_agent(uid):
            actor = Mock()
            actor.get_trainable_weights = Mock(return_value=[])
            return ActorCriticLOLAAgent(actor, None, 0, 0, agent_uid=uid)

        lola_agents = [create_lola_agent(uid) for uid in range(3)]
        add_opponent_actors(lola_agents)
        for agent1 in lola_agents:
            for agent2 in lola_agents:
                if agent1 != agent2:
                    self.assertEqual(agent1.opponent_actors[agent2.agent_uid], agent2.actor)
                else:
                    self.assertFalse(agent2.agent_uid in agent1.opponent_actors.keys())


###################
#  USEFUL CLASSES #
###################


class SuperSimplePolicy:

    def __init__(self, weights):
        self.trainable_weights = [weights]

    def __call__(self, state, training=None):
        l = state @ self.trainable_weights[0]
        return tf.cast(l, dtype=tf.float32)


class SimplePolicy:

    def __init__(self):
        self.trainable_weights = [tf.Variable([[1, 2, 3], [2, 3, 4], [4, 5, 6]], dtype=tf.dtypes.float32)]

    def __call__(self, ins, training=None):
        return ins @ self.trainable_weights[0]


if __name__ == '__main__':
    unittest.main()
