from agents import AgentFactory
from games import GameFactory
from globals import Time
from tournament.rl_tournament import TournamentFactory
from utils.utils import create_dir_if_not_exists, initialize_csv_file_if_empty_or_non_existing


class CallbackFactory:

    @staticmethod
    def create_tournament_callbacks(game_name, **callback_config):
        if game_name.upper() == "DIPLOMACY":
            return TournamentCallbackList([DiplomacyTournamentCallback(**callback_config)])
        else:
            return TournamentCallbackList([TournamentCallback()])

    @staticmethod
    def create_game_callbacks(game_name, **callback_config):
        if game_name.upper() == "DIPLOMACY":
            return GameCallbackList([DiplomacyGameCallback(**callback_config)])
        else:
            return GameCallbackList([GameCallback()])


###################################################################
#                   Tournament Callbacks                          #
###################################################################


class TournamentCallback:

    def on_episode_end(self, **kwargs):
        pass

    def on_episode_start(self, **kwargs):
        pass

    def on_step_end(self, **kwargs):
        pass


class TournamentCallbackList:

    def __init__(self, callbacks):
        self.callbacks = callbacks

    def on_step_end(self, **kwargs):
        for callback in self.callbacks:
            callback.on_step_end(**kwargs)

    def on_episode_end(self, **kwargs):
        for callback in self.callbacks:
            callback.on_episode_end(**kwargs)

    def on_episode_start(self, **kwargs):
        for callback in self.callbacks:
            callback.on_episode_start(**kwargs)

    def on_training_end(self, **kwargs):
        for callback in self.callbacks:
            callback.on_training_end(**kwargs)


class DiplomacyTournamentCallback(TournamentCallback):
    """
    Callback used in a Diplomacy tournament
    """

    def __init__(self, save_checkpoints_every_n_episodes, run_test_tournament_every_n_episode,
                 test_tournament_configs, sub_dir='', **kwargs):
        """
        :param save_checkpoints_every_n_episodes: How often model checkpoints should be saved
        :param run_test_tournament_every_n_episode: How often a test tournament should be run
        :param: test_tournament_config:
        """
        self.save_every_n_episodes = save_checkpoints_every_n_episodes
        self.run_test_tournament_every_n_episode = run_test_tournament_every_n_episode
        self.test_tournament_configs = test_tournament_configs
        if sub_dir:
            self.sub_dir = sub_dir
        else:
            self.sub_dir = Time.START_TIME

    def on_episode_end(self, **kwargs):
        episode_nr = kwargs["episode_nr"]
        agents = kwargs["agents"]
        if episode_nr % self.run_test_tournament_every_n_episode == 0:
            self.run_test_tournaments(agents, episode_nr)
        if episode_nr % self.save_every_n_episodes == 0:
            agents[0].actor.policy.save_weights(
                f"./checkpoints/{self.sub_dir}/policy/checkpoint-{kwargs['episode_nr']}")
            agents[0].critic.value_function.save_weights(
                f"./checkpoints/{self.sub_dir}/value_function/checkpoint-{kwargs['episode_nr']}")

    def run_test_tournaments(self, agents, episode_nr):
        """
        Run a test tournament.
        :param agents:
        :return:
        """
        for test_tournament_config in self.test_tournament_configs:
            self.run_test_tournament(test_tournament_config, episode_nr, agents)

    def run_test_tournament(self, test_tournament_config, episode_nr, agents):
        agent_configs = test_tournament_config["agents"]
        random_agents = self.create_random_agents(agent_configs)
        non_random_agents = self.find_non_random_agents(agents, agent_configs)
        agents = random_agents + non_random_agents
        game_callback = CallbackFactory.create_game_callbacks("DIPLOMACY",
                                                              file_name=f"""{test_tournament_config["name"]}-episode-{episode_nr}-test_log.csv""")
        game = GameFactory.create_game(agents, game_callback, **test_tournament_config["game"])
        callback_list = CallbackFactory.create_tournament_callbacks("DEFAULT")
        tournament = TournamentFactory.create_tournament("DEFAULT", game, callback_list,
                                                         **test_tournament_config["tournament"])
        tournament.play_games()

    def find_non_random_agents(self, agents, agent_configs):
        non_random_agents = []
        for agent_config in agent_configs:
            for agent in agents:
                if agent.name == agent_config["agent_name"]:
                    pnet = agent.actor.policy
                    for i in range(agent_config["exemplars"]):
                        non_random_agents.append(AgentFactory.create_agent(agent_type="A2C",
                                                                           pnet=pnet,
                                                                           vnet=None,
                                                                           gamma=1.0,
                                                                           **agent_config))
        return non_random_agents

    def create_random_agents(self, agent_configs):
        random_agents = []
        for agent_config in agent_configs:
            if agent_config["agent_name"] == "RANDOM":
                random_agents += [AgentFactory.create_agent(agent_type="A2C",
                                                            critic_type=None,
                                                            pnet=None,
                                                            vnet=None,
                                                            gamma=1.,
                                                            **agent_config) for _ in range(agent_config["exemplars"])]
        return random_agents


##############################################################
#                       Game Callbacks                       #
##############################################################

class GameCallbackList:

    def __init__(self, callbacks):
        self.callbacks = callbacks

    def on_step_end(self, **kwargs):
        for callback in self.callbacks:
            callback.on_step_end(**kwargs)


class GameCallback:

    def on_step_end(self, **kwargs):
        pass


class DiplomacyGameCallback(GameCallback):

    def __init__(self, file_name, sub_dir='', **kwargs):
        if sub_dir:
            self.log_file_path = f"logs/{sub_dir}/{file_name}"
        else:
            self.log_file_path = f"logs/{Time.START_TIME}/{file_name}"

    def on_step_end(self, **kwargs):
        create_dir_if_not_exists(self.log_file_path)
        initialize_csv_file_if_empty_or_non_existing(self.log_file_path, "step;phase;player_name;power;orders;reward")
        phase = kwargs["phase"]
        rewards = kwargs["rewards"]
        players_to_power_names = kwargs["players_to_power_names"]
        orders_pr_player = kwargs["orders"]
        step = kwargs["step"]
        with open(self.log_file_path, "a") as diplomacy_log:
            for player, orders in orders_pr_player.items():
                diplomacy_log.write(
                    f"\n{step};{phase};{player.name};{players_to_power_names[player]};{orders};{rewards[player]}")
