from diplomacy import Game as DiplomacyGameEngine
import numpy as np
import re
from games.dip.diplomacy_constants import DiplomacyConstantFactory
from collections import OrderedDict
from utils.utils import flip_dictionary
import random


class DiplomacyGame:
    """
    An interface between an agent and the game of diplomacy
    """

    def __init__(self, players, max_steps=40, map_name="standard", assign_powers_to_players_randomly=False,
                 callback_list=None):
        """
        :param players: A list of players
        :param max_steps: The maximum number of steps allowed in this game
        :param map_name: The name of the game map
        :param assign_powers_to_players_randomly: Whether to assign powers to players randomly. If True,
        a new random assignment is made every time the reset method is called.
        :param callback_list: A list of callbacks. The callbacks are called after each step.
        """
        self.__game_engine = DiplomacyGameEngine(map_name=map_name)
        if len(players) != len(self.__game_engine.powers.keys()):
            raise ValueError(f"The diplomacy game require {len(self.__game_engine.powers.keys())} players."
                             f"Only {len(players)} where given")
        self.__players = players
        self.__power_assignments = DiplomacyPowerAssignments(players, list(self.__game_engine.powers.keys()),
                                                             assign_powers_to_players_randomly)
        self.__diplomacy_constants = DiplomacyConstantFactory.get_constants(map_name)
        self.__map_name = map_name
        self.__max_steps = max_steps
        self.__step_count = 0
        self.__callback_list = callback_list
        self.__location_types = {k.upper(): v for k, v in self.__game_engine.map.loc_type.items()}
        self.__current_board_state = self.__get_current_board_state()

    def get_available_actions(self, player):
        """
        Returns the actions available to the player, and the power the player is playing as.

        available_order_encodings: An encoding of each possible order. shape=[1,
        P, max available orders pr province, #Features]. P is one less than the number of provinces a power needs to win the game.
        Each order in the available_order_encodings is encoded as follows: 1. the index of the province
        where the order is given. 2. an integer representation of the order. 3. one-hot encoding
        of the order type. 4. one-hot encoding of the underlying order type. 5. one-hot
        encoding of the source province. 6. one-hot encoding of the destination province.
        7. one-hot encoding of the unit.

        power: The power to which the orders are available. shape=[1].
        :param player:
        :return: available order encoding and the power the player is playing as.
        """
        power = self.__power_assignments.get_power(player)
        power_encoding = self.__encode_power(power)
        orderable_locations = self.__game_engine.get_orderable_locations(power)
        available_order_encoding = self.__current_board_state.to_available_order_encoding(orderable_locations)
        return [available_order_encoding, power_encoding]

    def step(self, actions):
        """
        Applies the provided actions
        :param actions: A dictionary mapping players into actions. Each action is a list of integer
        representations of orders. shape=[#orders, 1]
        :return: The state of the game, the rewards the players got by applying the provided actions,
        a boolean value telling whether the game is finished.

        game_state: The information about the game state which is available to the players:
                        1. An encoding of the current board state.
                        2. An encoding of the board state from the previous move phase
                        3. List of orders, represented by integers, which was applied in the previous
                           move phase
                        4. An integer representation of the current season.
                        5. Current build numbers

        rewards: A dictionary mapping players into rewards

        is_finished: whether the game is finished or not
        """
        orders = self.__convert_actions_to_orders(actions)
        self.__apply_orders(orders)
        phase = self.__game_engine.phase
        self.__game_engine.process()
        self.__current_board_state = self.__get_current_board_state()
        self.__step_count += 1
        rewards = self.__get_rewards()
        if self.__callback_list:
            self.__callback_list.on_step_end(rewards=rewards, orders=orders, phase=phase,
                                             players_to_power_names=self.__power_assignments.get_players_to_power_names(),
                                             step=self.__step_count)
        return self.__get_state(), rewards, self.__is_finished()

    def reset(self):
        """
        Resets the game and returns the new game state.
        :return: The information about the game state which is available to the players:
                        1. An encoding of the current board state.
                        2. An encoding of the board state from the previous move phase
                        3. List of orders, represented by integers, which was applied in the previous
                           move phase
                        4. An integer representation of the current season.
                        5. Current build numbers
        """
        self.__game_engine = DiplomacyGameEngine(map_name=self.__map_name)
        self.__current_board_state = self.__get_current_board_state()
        self.__step_count = 0
        self.__power_assignments.assign_powers_to_players()
        return self.__get_state()

    def get_players(self):
        """
        Returns the players
        :return: the players
        """
        return self.__players

    def get_winning_center_count(self):
        """
        Returns the number of supply centers needed to win the game
        :return: the number of supply centers needed to win the game
        """
        return self.__diplomacy_constants.WINNING_CENTER_COUNT

    def get_max_orders_pr_location(self):
        """
        Returns the maximum number of orders pr. location
        :return: the maximum number of orders pr. location
        """
        return self.__diplomacy_constants.MAX_ORDERS_PR_LOCATION

    def __convert_actions_to_orders(self, actions):
        return {player: self.__order_indices_to_orders(player_orders) for player, player_orders in actions.items()}

    def __order_indices_to_orders(self, order_indices):
        return [self.__diplomacy_constants.ORDER_VOCABULARY[int(order_idx)] for order_idx in order_indices if
                self.__is_valid_order(int(order_idx))]

    def __is_valid_order(self, order_idx):
        return 0 <= order_idx < len(self.__diplomacy_constants.ORDER_VOCABULARY)

    def __apply_orders(self, orders):
        for player, player_orders in orders.items():
            self.__apply_orders_for_player(player, player_orders)

    def __apply_orders_for_player(self, player, player_orders):
        power_name = self.__power_assignments.get_power(player)
        self.__game_engine.set_orders(power_name, player_orders)

    def __get_rewards(self):
        if self.__game_engine.is_game_done:
            rewards = self.__get_single_winner_reward()
        elif self.__has_reached_max_steps():
            rewards = self.__get_sum_of_squares_rewards()
        else:
            rewards = self.__get_default_rewards()
        return rewards

    def __get_single_winner_reward(self):
        rewards = self.__get_default_rewards()
        rewards[self.__power_assignments.get_player(self.__get_winning_power())] = 1
        return rewards

    def __has_reached_max_steps(self):
        return self.__step_count >= self.__max_steps

    def __get_sum_of_squares_rewards(self):
        centers = self.__game_engine.get_centers()
        denominator = sum(len(centers[self.__power_assignments.get_power(player)]) ** 2 for player in self.__players)
        return {player: (len(centers[self.__power_assignments.get_power(player)]) ** 2) / denominator for player in
                self.__players}

    def __get_default_rewards(self):
        return {player: 0 for player in self.__players}

    def __get_winning_power(self):
        centers = self.__game_engine.get_centers()
        for power, center_count in centers.items():
            has_won = len(center_count) >= self.__diplomacy_constants.WINNING_CENTER_COUNT
            if has_won:
                return power

    def __get_state(self):
        prev_movement_phase = PrevMovementPhase(self.__location_types,
                                                self.__get_prev_movement_phase(self.__game_engine.get_phase_history()),
                                                self.__diplomacy_constants)
        return [np.expand_dims(self.__current_board_state.to_board_state_encoding(), axis=0),
                np.expand_dims(prev_movement_phase.to_board_state_encoding(), axis=0),
                np.expand_dims(prev_movement_phase.to_order_encoding(), axis=0).astype(np.int64),
                np.expand_dims(self.__current_board_state.to_season_encoding(), axis=0).astype(np.int64),
                np.expand_dims(self.__current_board_state.to_build_number_encoding(), axis=0).astype(np.int64)]

    def __get_current_board_state(self):
        return StateWithAvailableOrders(self.__location_types, self.__game_engine.get_state(),
                                        self.__game_engine.get_all_possible_orders(),
                                        self.__diplomacy_constants)

    def __is_finished(self):
        return self.__has_reached_max_steps() or self.__game_engine.is_game_done

    def __encode_power(self, power):
        return np.expand_dims(self.__diplomacy_constants.POWER.index(power), axis=0).astype(np.int64)

    def __get_prev_movement_phase(self, phase_history):
        if self.__has_prev_movement_phase(phase_history):
            return self.__find_latest_movement_phase_data(phase_history)
        else:
            return DefaultPhaseData()

    def __has_prev_movement_phase(self, phase_history):
        return len(phase_history) >= 1

    def __find_latest_movement_phase_data(self, phase_history):
        for phase in reversed(phase_history):
            is_movement_phase = self.__parse_phase_type(phase.name) == "M"
            if is_movement_phase:
                return phase

    def __parse_phase_type(self, phase_name):
        return phase_name[-1]


class DefaultPhaseData:

    def __init__(self):
        self.orders = {}
        self.state = {
            "units": {},
            "builds": {},
            "centers": {},
            "name": None
        }


class DiplomacyPowerAssignments:
    """
    Class holding information about which power each player is playing as
    """

    def __init__(self, players, powers, assign_powers_to_players_randomly):
        self.__players = players
        self.__powers = powers
        self.__assign_powers_to_players_randomly = assign_powers_to_players_randomly
        self.__players_to_power_names = None
        self.__power_names_to_players = None
        self.assign_powers_to_players()

    def get_players(self):
        """
        Return all players
        :return: all players
        """
        return self.__players

    def get_player(self, power_name):
        """
        Return player
        :param power_name:
        :return:
        """
        return self.__power_names_to_players[power_name]

    def get_power(self, player):
        return self.__players_to_power_names[player]

    def get_players_to_power_names(self):
        return self.__players_to_power_names

    def assign_powers_to_players(self):
        if self.__assign_powers_to_players_randomly:
            self.__randomly_assign_power_names()
        else:
            self.__deterministically_assign_powers_to_players()

    def __randomly_assign_power_names(self):
        random.shuffle(self.__powers)
        self.__players_to_power_names = {self.__players[i]: self.__powers[i] for i in range(len(self.__players))}
        self.__power_names_to_players = flip_dictionary(self.__players_to_power_names)

    def __deterministically_assign_powers_to_players(self):
        self.__players_to_power_names = {self.__players[i]: self.__powers[i] for i in range(len(self.__players))}
        self.__power_names_to_players = flip_dictionary(self.__players_to_power_names)


class State:
    """
    Represents a state in a game of Diplomacy
    """

    def __init__(self, location_types, game_state, diplomacy_constants):
        """
        :param location_types: A dictionary mapping locations into location types (land, water, coast)
        :param game_state: A game state
        :param diplomacy_constants: Diplomacy constants
        """
        # Switzerland is part of the game_map, but doesn't have any function in the game.
        # Switzerland is therefore ignored
        locations = {location.upper(): Location(location.upper(), location_types[location], diplomacy_constants)
                     for location in diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS if location != "SWI"}
        self._diplomacy_constants = diplomacy_constants
        # adapted from https://stackoverflow.com/questions/14659872/sort-a-dictionary-using-index-position-of-keys-in-a-base-list
        self._locations = OrderedDict(
            sorted(locations.items(), key=lambda t: self._diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS.index(t[0])))
        self.__season = None
        self.__build_state(game_state)

    def to_board_state_encoding(self):
        """
        Returns a one-hot encoding of the board state.
        :return:  a one-hot encoding of the board state. shape=[#locations,
        encoding length] .
        The one-hot encoding contains information about:
        1. the unit type
        2. the unit power
        3. whether a unit can be built in the location
        4. whether the unit in the location can can be removed
        5. If a unit is dislodged, which unit type is it
        6. If a unit is dislodged, which power owned the unit
        7. Which type the area is (land, water, coast)
        8. If the location is a supply center, which power owns the supply center
        """
        return np.array([location.to_location_state_encoding() for location in self._locations.values()])

    def to_season_encoding(self):
        """
        Returns an encoding of the season of the game.
        :return: an encoding of the season of the game. The encoding is an integer
        representation of the season. shape=[#locations, 1]
        """
        season_index = self._diplomacy_constants.SEASON.index(self.__season)
        locations_count = self.__get_location_count()
        return np.array([[season_index] for _ in range(locations_count)])

    def to_build_number_encoding(self):
        """
        Returns the build numbers associated with each location
        :return: the build numbers associated with each location. shape=[#locations, 1]
        """
        return np.array([[location.get_build_number()] for location in self._locations.values()])

    def __get_location_count(self):
        return len(self._locations.keys())

    def __build_state(self, game_state):
        self.__season = self.__parse_season(game_state["name"])
        self.__add_unit_info_to_locations(game_state)
        self.__add_build_information_to_locations(game_state)
        self.__add_supply_center_owner_to_locations(game_state)

    def __add_unit_info_to_locations(self, game_state):
        for power, units in game_state["units"].items():
            for unit_str in units:
                self.__add_unit_info_to_location(game_state, power, unit_str)

    def __add_unit_info_to_location(self, game_info, power, unit_str):
        dislodged, unit, location_name = self.__parse_unit_str(unit_str)
        location = self._locations[location_name]
        location.set_unit(unit, dislodged)
        location.set_power(power, dislodged)
        location.set_removable(game_info["builds"][power]["count"] < 0)
        if location.has_parent_location():
            self.__add_unit_info_to_location(game_info, power, unit_str[:-3])

    def __add_build_information_to_locations(self, state):
        for power, builds in state["builds"].items():
            for home in builds["homes"]:
                location = self._locations[home]
                location.set_build_number(builds["count"])
                location.set_buildable(builds["count"] > 0)

    def __add_supply_center_owner_to_locations(self, state):
        for power, centers in state["centers"].items():
            for center in centers:
                location = self._locations[center]
                location.set_supply_center_owner(power)

    def __parse_unit_str(self, unit_str):
        dislodged = False
        if unit_str[0] == "*":
            dislodged = True
            unit_str = unit_str[1:]
        match = re.fullmatch(f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION})", unit_str)
        return dislodged, match.group("unit"), match.group("location")

    def __parse_season(self, phase_name):
        if phase_name and phase_name[0] in ["S", "F", "W"]:
            return phase_name[0]


class StateWithAvailableOrders(State):
    """
    Represents a state of a game of Diplomacy. Includes information
    about available orders
    """

    def __init__(self, location_types, game_state, available_orders, diplomacy_constants):
        super().__init__(location_types, game_state, diplomacy_constants)
        self.__default_available_order_encoding = Location(None, None,
                                                           diplomacy_constants).to_available_order_encoding()
        self.__add_available_orders_to_locations(available_orders)

    def to_available_order_encoding(self, orderable_locations):
        """
        Returns an encoding of available orders
        :param orderable_locations: A list of location where orders can be made
        :return: an encoding of available orders.  shape=[1,
        P, max available orders pr province, #Features]. P is one less than the number of provinces a power needs to win the game.
        Each order in the available_order_encodings is encoded as follows: 1. the index of the province
        where the order is given. 2. an integer representation of the order. 3. one-hot encoding
        of the order type. 4. one-hot encoding of the underlying order type. 5. one-hot
        encoding of the source province. 6. one-hot encoding of the destination province.
        7. one-hot encoding of the unit.
        """
        available_order_encoding = [location.to_available_order_encoding() for location in self._locations.values() if
                                    location.has_available_orders() and location.name in orderable_locations]
        available_order_encoding_padding = [self.__default_available_order_encoding for _ in
                                            range((self._diplomacy_constants.WINNING_CENTER_COUNT - 1) - len(
                                                available_order_encoding))]
        return np.expand_dims(np.array(available_order_encoding + available_order_encoding_padding), axis=0)

    def __add_available_orders_to_locations(self, possible_orders):
        for location_name, orders in possible_orders.items():
            self.__add_available_orders_to_location(location_name, orders)

    def __add_available_orders_to_location(self, location_name, orders):
        for order_str in orders:
            try:
                order = OrderParser.parse_order_str(order_str)
                location = self._locations[location_name]
                location.add_available_order(order)
            except ValueError:
                print(f"Order {order_str} was not added to the list of possible orders")


class PrevMovementPhase(State):
    """
    Represents the previous movement phase in a game of Diplomacy. Contains information
    about the state in the previous move phase, and the orders which where given
    in that phase.
    """

    def __init__(self, location_types, prev_movement_phase_data, diplomacy_constants):
        """
        :param location_types: A dictionary mapping locations into location types (land, water, coast)
        :param prev_movement_phase_data: Data from the previous movement phase (from the game engine)
        :param diplomacy_constants: Diplomacy constants
        """
        super().__init__(location_types, prev_movement_phase_data.state, diplomacy_constants)
        self.__build_phase(prev_movement_phase_data)

    def to_order_encoding(self):
        """
        Returns an encoding of the orders given in the previous movement phase
        :return: an encoding of the orders given in the previous movement phase. Each order
        is represented by an integer. shape=[#locations, 1]
        """
        return np.array([location.to_order_encoding() for location in self._locations.values()])

    def __build_phase(self, phase_data):
        self.__add_previous_orders(phase_data)

    def __add_previous_orders(self, phase_data):
        for power, orders in phase_data.orders.items():
            for order_str in orders:
                order = OrderParser.parse_order_str(order_str)
                location = self._locations[order.location_name]
                location.set_order(order)


class OrderParser:

    @staticmethod
    def parse_order_str(order_str):
        """
        Parses the order_str and returns an Order object containing the information
        in the order_str
        :param order_str: A string representation of an order
        :return: A Order object containing the information in the order_str
        """
        hold_order = OrderParser.parse_hold_order(order_str)
        if hold_order:
            return hold_order
        move_order = OrderParser.parse_move_order(order_str)
        if move_order:
            return move_order
        support_order = OrderParser.parse_support_order(order_str)
        if support_order:
            return support_order
        convoy_order = OrderParser.parse_convoy_order(order_str)
        if convoy_order:
            return convoy_order
        build_order = OrderParser.parse_build_order(order_str)
        if build_order:
            return build_order
        waive_order = OrderParser.parse_waive_order(order_str)
        if waive_order:
            return waive_order
        disband_order = OrderParser.parse_disband_order(order_str)
        if disband_order:
            return disband_order
        retreat_order = OrderParser.parse_retreat_order(order_str)
        if retreat_order:
            return retreat_order
        raise ValueError(f"Not a valid order: {order_str}")

    @staticmethod
    def parse_hold_order(order_str):
        try:
            hold_order = re.fullmatch(
                f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION}){DipRegEx.SPACE}"
                f"{DipRegEx.HOLD_ORDER}", order_str)
            return Order(order_str=order_str, location_name=hold_order.group("location"),
                         unit=hold_order.group("unit"), order_type=OrderTypes.HOLD,
                         src_location=hold_order.group("location"),
                         dst_location=hold_order.group("location"),
                         underlying_order_type=OrderTypes.HOLD)
        except AttributeError:
            return None

    @staticmethod
    def parse_move_order(order_str):
        try:
            move_order = re.match(f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION})"
                                  f"{DipRegEx.SPACE}[-]{DipRegEx.SPACE}(?P<dst_location>{DipRegEx.LOCATION})",
                                  order_str)
            return Order(order_str=order_str, location_name=move_order.group("location"),
                         unit=move_order.group("unit"), order_type=OrderTypes.MOVE,
                         src_location=move_order.group("location"),
                         dst_location=move_order.group("dst_location"),
                         underlying_order_type=OrderTypes.MOVE)
        except AttributeError:
            return None

    @staticmethod
    def parse_support_order(order_str):
        try:
            support_order = re.fullmatch(
                f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION}){DipRegEx.SPACE}{DipRegEx.SUPPORT_ORDER}"
                f"{DipRegEx.SPACE}(?P<underlying_order_unit>{DipRegEx.UNIT}){DipRegEx.SPACE}((?P<src_location>{DipRegEx.LOCATION})|(?P<src_location_alt>{DipRegEx.LOCATION})"
                f"{DipRegEx.SPACE}[-]{DipRegEx.SPACE}(?P<dst_location>{DipRegEx.LOCATION}))", order_str)
            if support_order.group("src_location_alt"):
                return Order(order_str=order_str,
                             location_name=support_order.group("location"),
                             unit=support_order.group("underlying_order_unit"),
                             order_type=OrderTypes.SUPPORT,
                             src_location=support_order.group("src_location_alt"),
                             dst_location=support_order.group("dst_location"),
                             underlying_order_type=OrderTypes.MOVE)
            else:
                return Order(order_str=order_str,
                             location_name=support_order.group("location"),
                             order_type=OrderTypes.SUPPORT,
                             src_location=support_order.group("src_location"),
                             dst_location=support_order.group("src_location"),
                             underlying_order_type=OrderTypes.HOLD,
                             unit=support_order.group("underlying_order_unit"))
        except AttributeError:
            return None

    @staticmethod
    def parse_convoy_order(order_str):
        try:
            convoy_order = re.fullmatch(
                f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION}){DipRegEx.SPACE}{DipRegEx.CONVOY_ORDER}"
                f"{DipRegEx.SPACE}(?P<underlying_order_unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<src_location>{DipRegEx.LOCATION}){DipRegEx.SPACE}[-]{DipRegEx.SPACE}"
                f"(?P<dst_location>{DipRegEx.LOCATION})", order_str)
            return Order(order_str=order_str, location_name=convoy_order.group("location"),
                         order_type=OrderTypes.CONVOY,
                         src_location=convoy_order.group("src_location"),
                         dst_location=convoy_order.group("dst_location"),
                         underlying_order_type=OrderTypes.MOVE,
                         unit=convoy_order.group("underlying_order_unit"))
        except AttributeError:
            return None

    @staticmethod
    def parse_build_order(order_str):
        try:
            build_order = re.fullmatch(f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION})"
                                       f"{DipRegEx.SPACE}{DipRegEx.BUILD_ORDER}", order_str)
            return Order(order_str=order_str, location_name=build_order.group("location"),
                         unit=build_order.group("unit"), order_type=OrderTypes.BUILD,
                         src_location=build_order.group("location"),
                         dst_location=build_order.group("location"),
                         underlying_order_type=OrderTypes.BUILD)
        except AttributeError:
            return None

    @staticmethod
    def parse_disband_order(order_str):
        try:
            disband_order = re.fullmatch(f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION})"
                                         f"{DipRegEx.SPACE}{DipRegEx.DISBAND_ORDER}", order_str)
            return Order(order_str=order_str, location_name=disband_order.group("location"),
                         unit=disband_order.group("unit"), order_type=OrderTypes.DISBAND,
                         dst_location=disband_order.group("location"),
                         src_location=disband_order.group("location"),
                         underlying_order_type=OrderTypes.DISBAND)
        except AttributeError:
            return None

    @staticmethod
    def parse_retreat_order(order_str):
        try:
            retreat_order = re.match(f"(?P<unit>{DipRegEx.UNIT}){DipRegEx.SPACE}(?P<location>{DipRegEx.LOCATION})"
                                     f"{DipRegEx.SPACE}{DipRegEx.RETREAT_ORDER}{DipRegEx.SPACE}"
                                     f"(?P<dst_location>{DipRegEx.LOCATION})",
                                     order_str)
            return Order(order_str=order_str, location_name=retreat_order.group("location"),
                         unit=retreat_order.group("unit"), order_type=OrderTypes.RETREAT,
                         src_location=retreat_order.group("location"),
                         dst_location=retreat_order.group("dst_location"),
                         underlying_order_type=OrderTypes.RETREAT)
        except AttributeError:
            return None

    @staticmethod
    def parse_waive_order(order_str):
        if order_str == "WAIVE":
            return Order()
        else:
            return None


class Location:
    """
    A location on a Diplomacy map
    """

    def __init__(self, name, area_type, diplomacy_constants):
        self.name = name
        self.diplomacy_constants = diplomacy_constants
        self.location_index = self.diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS_WITH_NONE.index(name)
        self.unit_type = None
        self.unit_power = None
        self.buildable = False
        self.build_number = 0
        self.removable = False
        self.dislodged_unit = None
        self.dislodged_unit_power = None
        self.area_type = area_type
        self.supply_center_owner = None
        self.order = Order()
        self.available_orders = []
        self.default_order_encoding = self.__encode_single_available_order(Order())

    def set_unit(self, unit, dislodged=False):
        if dislodged:
            self.dislodged_unit = unit
        else:
            self.unit_type = unit

    def set_power(self, power, dislodged=False):
        if dislodged:
            self.dislodged_unit_power = power
        else:
            self.unit_power = power

    def set_removable(self, removable):
        self.removable = removable

    def set_buildable(self, buildable):
        self.buildable = buildable

    def set_build_number(self, build_number):
        self.build_number = build_number

    def get_build_number(self):
        return self.build_number

    def set_supply_center_owner(self, sc_owner):
        self.supply_center_owner = sc_owner

    def has_parent_location(self):
        return "/" in self.name

    def get_owner(self):
        if self.unit_power is not None:
            return self.unit_power
        else:
            return self.supply_center_owner

    def set_order(self, order):
        self.order = order

    def add_available_order(self, order):
        self.available_orders.append(order)

    def has_available_orders(self):
        return len(self.available_orders) > 0

    def to_location_state_encoding(self):
        """
        Returns a one-hot encoding of the state of the location
        :return: a one-hot encoding of the state of the location. shape=[1, #features]
        The one-hot encoding contains information about:
        1. the unit type
        2. the unit power
        3. whether a unit can be built in the location
        4. whether the unit in the location can can be removed
        5. If a unit is dislodged, which unit type is it
        6. If a unit is dislodged, which power owned the unit
        7. Which type the area is (land, water, coast)
        8. If the location is a supply center, which power owns the supply center
        """
        unit_type = self.__encode_unit_array(self.unit_type)
        unit_power = self.__encode_power_array(self.unit_power)
        buildable = np.array([int(self.buildable)])
        removable = np.array([int(self.removable)])
        dislodged_unit_type = self.__encode_unit_array(self.dislodged_unit)
        dislodged_unit_power = self.__encode_power_array(self.dislodged_unit_power)
        area_type = self.__encode_area_type()
        supply_center_owner = self.__encode_power_array(self.supply_center_owner)
        return np.concatenate((unit_type, unit_power, buildable, removable, dislodged_unit_type,
                               dislodged_unit_power, area_type, supply_center_owner))

    def to_order_encoding(self):
        """
        Return an integer representation of the order which was given in this location
        :return: an integer representation of the order which was given in this location.
        """
        return self.__convert_to_order_index(self.order.order_str)

    def to_available_order_encoding(self):
        """
        Return an encoding of the orders which is available in this location
        :return: an encoding of the orders which is available in this location.
        shape=[max available orders pr province, #Features].
         Each order in the available_order_encodings is encoded as follows: 1. the index of the province
        where the order is given. 2. an integer representation of the order. 3. one-hot encoding
        of the order type. 4. one-hot encoding of the underlying order type. 5. one-hot
        encoding of the source province. 6. one-hot encoding of the destination province.
        7. one-hot encoding of the unit.
        """
        available_order_encoding = [self.__encode_single_available_order(order) for order in
                                    self.available_orders]
        available_order_padding = [self.default_order_encoding for _ in
                                   range(self.diplomacy_constants.MAX_ORDERS_PR_LOCATION - len(
                                       self.available_orders))]
        return np.array(available_order_encoding + available_order_padding)

    def __encode_single_available_order(self, order):
        order_index = self.__convert_to_order_index(order.order_str)
        location_index = np.array([self.location_index])
        order_type = self.__encode_order_type(order.order_type)
        unit = self.__encode_unit_array(order.unit)
        underlying_order_type = self.__encode_order_type(order.underlying_order_type)
        src_location = self.__encode_location(order.src_location)
        dst_location = self.__encode_location(order.dst_location)
        return np.concatenate(
            (location_index, order_index, order_type, underlying_order_type, src_location, dst_location,
             unit))

    def __convert_to_order_index(self, order_str):
        if order_str is None:
            return np.array([len(self.diplomacy_constants.ORDER_VOCABULARY)])
        else:
            return np.array([self.diplomacy_constants.ORDER_VOCABULARY.index(order_str)])

    def __encode_power_array(self, power):
        power_array = np.zeros(len(self.diplomacy_constants.POWER))
        power_array[self.diplomacy_constants.POWER.index(power)] = 1
        return power_array

    def __encode_unit_array(self, unit):
        unit_array = np.zeros(len(self.diplomacy_constants.UNIT_TYPE))
        unit_array[self.diplomacy_constants.UNIT_TYPE.index(unit)] = 1
        return unit_array

    def __encode_area_type(self):
        area_array = np.zeros(len(self.diplomacy_constants.AREA_TYPE))
        area_array[self.diplomacy_constants.AREA_TYPE.index(self.area_type)] = 1
        return area_array

    def __encode_order_type(self, order_type):
        order_array = np.zeros(len(self.diplomacy_constants.ORDER_TYPE))
        order_array[self.diplomacy_constants.ORDER_TYPE.index(order_type)] = 1
        return order_array

    def __encode_location(self, location):
        location_array = np.zeros(len(self.diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS_WITH_NONE))
        location_array[self.diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS_WITH_NONE.index(location)] = 1
        return location_array


class Order:
    """
    An order in a game of Diplomacy
    """

    def __init__(self, order_str=None, location_name=None, unit=None, order_type=None, src_location=None,
                 dst_location=None, underlying_order_type=None):
        self.order_str = order_str
        self.location_name = location_name
        self.unit = unit
        self.order_type = order_type
        self.src_location = src_location
        self.dst_location = dst_location
        self.underlying_order_type = underlying_order_type


class DipRegEx:
    UNIT = "[AF]"
    LOCATION = "([A-Z]{3}|[A-Z]{3}[/][A-Z]{2})"
    SPACE = "[\u0020]"
    HOLD_ORDER = "[H]"
    SUPPORT_ORDER = "[S]"
    CONVOY_ORDER = "[C]"
    BUILD_ORDER = "[B]"
    DISBAND_ORDER = "[D]"
    RETREAT_ORDER = "[R]"


class OrderTypes:
    MOVE = "MOVE"
    SUPPORT = "SUPPORT"
    HOLD = "HOLD"
    CONVOY = "CONVOY"
    RETREAT = "RETREAT"
    DISBAND = "DISBAND"
    BUILD = "BUILD"


def get_diplomacy_map_as_adjacency_matrix(map_name="standard"):
    """
    Returns adjacency matrix representation of the map with the provided name
    :param map_name: The name of the map
    :return: adjacency matrix representation of the map with the provided name
    """
    game = DiplomacyGameEngine(map_name=map_name)
    abuts = game.map.loc_abut
    diplomacy_constants = DiplomacyConstantFactory.get_constants(map_name)
    num_nodes = len(diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS)
    adjacency_matrix = np.zeros(shape=[num_nodes, num_nodes])
    for u, vs in abuts.items():
        for v in vs:
            if u != "SWI" and v != "SWI":
                adjacency_matrix[
                    diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS.index(u.upper()),
                    diplomacy_constants.TOPOLOGICAL_SORTED_LOCATIONS.index(v.upper())] = 1
    adjacency_matrix = np.logical_or(adjacency_matrix, adjacency_matrix.T)
    return adjacency_matrix
