import unittest
from unittest.mock import Mock

import numpy as np

from games.dip.dip import DiplomacyGame, get_diplomacy_map_as_adjacency_matrix, OrderParser, State, \
    StateWithAvailableOrders, PrevMovementPhase, Location
from utils.utils import flip_dictionary


class TestDiplomacyGame(unittest.TestCase):

    def setUp(self):
        self.diplomacy_game = DiplomacyGame([f"Player{i}" for i in range(7)], max_steps=10,
                                            map_name="pure")

    def test_initialization_with_invalid_number_of_players(self):
        self.assertRaises(ValueError, DiplomacyGame, [f"player{i}" for i in range(6)])
        self.assertRaises(ValueError, DiplomacyGame, [f"player{i}" for i in range(10)])
        self.assertRaises(ValueError, DiplomacyGame, [f"player{i}" for i in range(10)],
                          map_name="standard_france_austria")
        self.assertRaises(ValueError, DiplomacyGame, [f"player{i}" for i in range(1)],
                          map_name="standard_france_austria")

    def test_available_actions_correct_output_size(self):
        expected_encoding_length = 34
        available_actions, power = self.diplomacy_game.get_available_actions("Player1")
        self.assertEqual((1, self.diplomacy_game.get_winning_center_count() - 1,
                          self.diplomacy_game.get_max_orders_pr_location(),
                          expected_encoding_length), available_actions.shape)
        self.assertEqual((1,), power.shape)

    def test_step_correct_output_size(self):
        orders = {"Player1": [], "Player2": []}
        state, rewards, is_finished = self.diplomacy_game.step(orders)
        self.assertFalse(is_finished)
        self.assertEqual(7, len(rewards.keys()))
        self.assertEqual(5, len(state))

    def test_reset_correct_output_size(self):
        state = self.diplomacy_game.reset()
        self.assertEqual(5, len(state))

    def test_diplomacy_get_players(self):
        self.assertEqual(7, len(self.diplomacy_game.get_players()))
        for i in range(7):
            self.assertEqual(f"Player{i}", self.diplomacy_game.get_players()[i])

    def test_get_winning_center_count(self):
        self.assertEqual(4, self.diplomacy_game.get_winning_center_count())


class TestState(unittest.TestCase):

    def setUp(self):
        self.state = State(MockGameMap().loc_type, MockDiplomacyGameEngine().get_state(), MockDiplomacyConstants)

    def test_board_state_encoding(self):
        expected_board_state_encoding = np.array(
            [[0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0],
             [0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0],
             [1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1],
             [0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0],
             [1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0],
             [1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1],
             [1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1],
             [0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1],
             [0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0],
             [0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1],
             [0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0],
             [1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0]])

        board_state_encoding = self.state.to_board_state_encoding()
        np.testing.assert_equal(board_state_encoding, expected_board_state_encoding)

    def test_to_season_encoding(self):
        expected_season_encoding = np.array([[0] for _ in range(12)])
        actual_season_encoding = self.state.to_season_encoding()
        np.testing.assert_equal(actual_season_encoding, expected_season_encoding)

    def test_to_build_number_encodings(self):
        expected_build_number_encodings = [[0] for _ in range(5)] + [[-1]] + [[0] for _ in range(2)] + [[1]] + [[0] for
                                                                                                                _ in
                                                                                                                range(
                                                                                                                    3)]
        actual_build_number_encoding = self.state.to_build_number_encoding()
        np.testing.assert_equal(actual_build_number_encoding, expected_build_number_encodings)


class TestStateWithAvailableOrders(unittest.TestCase):

    def setUp(self):
        self.game_engine = MockDiplomacyGameEngine()
        self.state_with_available_orders = StateWithAvailableOrders(location_types=MockGameMap().loc_type,
                                                                    game_state=self.game_engine.get_state(),
                                                                    available_orders=self.game_engine.available_orders,
                                                                    diplomacy_constants=MockDiplomacyConstants)

    def test_to_available_order_encoding_correct_output_size(self):
        available_orders_encoding = self.state_with_available_orders.to_available_order_encoding(
            self.game_engine.get_orderable_locations("AUSTRIA"))
        self.assertEqual((1, MockDiplomacyConstants.WINNING_CENTER_COUNT - 1,
                          MockDiplomacyConstants.MAX_ORDERS_PR_LOCATION, 47), available_orders_encoding.shape)


class TestPrevMovementPhase(unittest.TestCase):

    def setUp(self):
        self.prev_movement_phase = PrevMovementPhase(MockGameMap().loc_type, MockGamePhaseData(),
                                                     MockDiplomacyConstants)

    def test_to_order_encoding(self):
        expected_order_encoding = [[6], [9], [7], [8]] + [[9] for _ in range(5)] + [[4], [5], [0]]
        actual_order_encoding = self.prev_movement_phase.to_order_encoding()
        np.testing.assert_equal(actual_order_encoding, expected_order_encoding)


class TestOrderParser(unittest.TestCase):

    def test_parsing_of_valid_hold_orders(self):
        order = OrderParser.parse_order_str("A LON H")
        self.assert_order_contain_param_values(order, unit="A", src_location="LON", dst_location="LON",
                                               location_name="LON", order_type="HOLD", underlying_order_type="HOLD")

        order = OrderParser.parse_order_str("F LON H")
        self.assert_order_contain_param_values(order, unit="F", src_location="LON", dst_location="LON",
                                               location_name="LON", order_type="HOLD", underlying_order_type="HOLD")

    def test_parsing_of_valid_move_orders(self):
        order = OrderParser.parse_order_str("A BER - PRU")
        self.assert_order_contain_param_values(order, location_name="BER", unit="A", order_type="MOVE",
                                               src_location="BER", dst_location="PRU", underlying_order_type="MOVE")

        order = OrderParser.parse_order_str("F BER - PRU")
        self.assert_order_contain_param_values(order, location_name="BER", unit="F", order_type="MOVE",
                                               src_location="BER", dst_location="PRU", underlying_order_type="MOVE")

        order = OrderParser.parse_order_str("A HOL - NWY VIA")
        self.assert_order_contain_param_values(order, location_name="HOL", unit="A",
                                               order_type="MOVE", src_location="HOL",
                                               dst_location="NWY",
                                               underlying_order_type="MOVE")

    def test_parsing_of_valid_support_orders(self):
        order = OrderParser.parse_order_str("A HOL S F KIE")
        self.assert_order_contain_param_values(order, location_name="HOL", unit="F",
                                               order_type="SUPPORT",
                                               src_location="KIE",
                                               dst_location="KIE",
                                               underlying_order_type="HOLD")

        order = OrderParser.parse_order_str("F HOL S A KIE")
        self.assert_order_contain_param_values(order, location_name="HOL", unit="A",
                                               order_type="SUPPORT",
                                               src_location="KIE",
                                               dst_location="KIE",
                                               underlying_order_type="HOLD")

        order = OrderParser.parse_order_str("A HOL S A CLY - BEL")
        self.assert_order_contain_param_values(order, location_name="HOL", unit="A",
                                               order_type="SUPPORT",
                                               src_location="CLY",
                                               dst_location="BEL",
                                               underlying_order_type="MOVE")

        order = OrderParser.parse_order_str("A HOL S F CLY - BEL")
        self.assert_order_contain_param_values(order, location_name="HOL", unit="F",
                                               order_type="SUPPORT",
                                               src_location="CLY",
                                               dst_location="BEL",
                                               underlying_order_type="MOVE")

    def test_parsing_of_valid_convoy_orders(self):
        order = OrderParser.parse_order_str("F ADR C A ALB - APU")
        self.assert_order_contain_param_values(order, location_name="ADR", unit="A",
                                               order_type="CONVOY",
                                               src_location="ALB",
                                               dst_location="APU",
                                               underlying_order_type="MOVE")

    def test_parsing_of_valid_disband_orders(self):
        order = OrderParser.parse_order_str("A NWY D")
        self.assert_order_contain_param_values(order, location_name="NWY", unit="A",
                                               order_type="DISBAND", src_location="NWY",
                                               dst_location="NWY",
                                               underlying_order_type="DISBAND")

    def test_parsing_of_valid_retreat_orders(self):
        order = OrderParser.parse_order_str("A NWY R SWE")
        self.assert_order_contain_param_values(order, location_name="NWY",
                                               unit="A",
                                               order_type="RETREAT",
                                               src_location="NWY",
                                               dst_location="SWE",
                                               underlying_order_type="RETREAT")

    def test_parsing_of_valid_build_orders(self):
        order = OrderParser.parse_order_str("A NWY B")
        self.assert_order_contain_param_values(order, location_name="NWY",
                                               unit="A",
                                               order_type="BUILD",
                                               src_location="NWY",
                                               dst_location="NWY",
                                               underlying_order_type="BUILD")

    def test_invalid_orders(self):
        orders = ["H IRO H", "A IRI-MAO-WAL"]
        for order in orders:
            self.assertRaises(ValueError, OrderParser.parse_order_str, order)

    def assert_order_contain_param_values(self, order,
                                          location_name=None,
                                          unit=None,
                                          order_type=None,
                                          src_location=None,
                                          dst_location=None,
                                          underlying_order_type=None):
        self.assertEqual(location_name, order.location_name)
        self.assertEqual(unit, order.unit)
        self.assertEqual(order_type, order.order_type),
        self.assertEqual(src_location, order.src_location),
        self.assertEqual(dst_location, order.dst_location),
        self.assertEqual(underlying_order_type, order.underlying_order_type)


class TestLocation(unittest.TestCase):

    def setUp(self):
        self.location = Location("PAR", "LAND", MockDiplomacyConstants)
        self.location.set_unit("A")
        self.location.set_power("FRANCE")
        self.location.set_supply_center_owner("FRANCE")
        order = Mock()
        order.order_str = "A VIE S A LON"
        order.unit = "A"
        order.order_type = "SUPPORT"
        order.src_location = "VIE"
        order.dst_location = "LON"
        order.underlying_order_type = "HOLD"
        self.location.set_order(order)
        self.location.add_available_order(order)

    def test_to_location_state_encoding(self):
        expected_location_state_encoding = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0]
        actual_location_state_encoding = self.location.to_location_state_encoding()
        np.testing.assert_equal(actual_location_state_encoding, expected_location_state_encoding)

        self.location.set_buildable(True)
        self.location.set_removable(True)
        expected_location_state_encoding = [1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0]
        actual_location_state_encoding = self.location.to_location_state_encoding()
        np.testing.assert_equal(actual_location_state_encoding, expected_location_state_encoding)

        self.location.set_unit("F", dislodged=True)
        self.location.set_power("AUSTRIA", dislodged=True)
        expected_location_state_encoding = [1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0]
        actual_location_state_encoding = self.location.to_location_state_encoding()
        np.testing.assert_equal(actual_location_state_encoding, expected_location_state_encoding)

    def test_to_order_encoding(self):
        expected_encoding = 3
        actual_encoding = self.location.to_order_encoding()
        self.assertEqual(actual_encoding, expected_encoding)

    def test_to_order_encoding_invalid_order(self):
        order = Mock()
        order.order_str = "MOS"
        self.location.set_order(order)
        self.assertRaises(ValueError, self.location.to_order_encoding)

    def test_to_available_order_encoding(self):
        expected_default_available_order_encoding = [5, 9, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1,
                                                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                                                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                                                     0, 0, 1]
        expected_available_order_encoding = [5, 3, 0, 0, 1, 0, 0, 0, 0,
                                             1, 0, 0, 0, 0, 0, 0,
                                             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0,
                                             0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                             1, 0, 0]
        expected_available_order_encoding_with_padding = [expected_available_order_encoding] + [expected_default_available_order_encoding for _ in range(4)]
        actual_available_order_encoding = self.location.to_available_order_encoding()
        np.testing.assert_equal(actual_available_order_encoding, expected_available_order_encoding_with_padding)


class TestGetDiplomacyMapAsAdjacencyMatrix(unittest.TestCase):

    def test_get_diplomacy_map_as_adjacency_matrix(self):
        adjacency_matrix = get_diplomacy_map_as_adjacency_matrix()
        self.assertEqual((81, 81), adjacency_matrix.shape)
        self.assertTrue(np.all((adjacency_matrix ^ adjacency_matrix.T) == 0))


###################
#  USEFUL CLASSES #
###################


class MockDiplomacyConstants:
    UNIT_TYPE = ["A", "F", None]
    AREA_TYPE = ["LAND", "WATER", "COAST"]
    ORDER_TYPE = ["HOLD", "MOVE", "SUPPORT", "BUILD", "DISBAND", "RETREAT", None]
    POWER = ["AUSTRIA", "ENGLAND", "FRANCE", None]
    MAX_ORDERS_PR_LOCATION = 5
    WINNING_CENTER_COUNT = 3
    SEASON = ["S", "F", "W", None]
    TOPOLOGICAL_SORTED_LOCATIONS = ["EDI", "LON", "LVP", "BRE", "STP", "PAR", "STP/SC", "MUN", "TUN", "SWI", "VIE",
                                    "TRI", "BUD"]
    TOPOLOGICAL_SORTED_LOCATIONS_WITH_NONE = ["EDI", "LON", "LVP", "BRE", "STP", "PAR", "STP/SC", "MUN", "TUN", "SWI",
                                              "VIE",
                                              "TRI", "BUD", None]

    ORDER_VOCABULARY = ["A BUD H",
                        "A BUD S A EDI - LVP",
                        "A VIE S A LON - TRI",
                        "A VIE S A LON",
                        "F VIE - TUN",
                        "F TRI - VIE",
                        "F EDI S F LON",
                        "A LVP S F VIE - TUN",
                        "F BRE C A STP/SC - PAR"]


class MockGameMap:

    def __init__(self):
        self.locs = ["EDI", "LON", "LVP", "BRE", "STP", "PAR", "STP/SC", "MUN", "TUN", "SWI", "VIE", "TRI", "BUD"]
        self.loc_type = {
            "BUD": "LAND",
            "VIE": "LAND",
            "TRI": "LAND",
            "EDI": "WATER",
            "LON": "WATER",
            "LVP": "WATER",
            "BRE": "WATER",
            "STP": "LAND",
            "PAR": "LAND",
            "MUN": "LAND",
            "STP/SC": "COAST",
            "TUN": "WATER",
            "SWI": "LAND"
        }


class MockDiplomacyGameEngine:

    def __init__(self):
        self.map = MockGameMap()
        self.phase = "SPRING 1901 MOVEMENT"
        self.is_game_done = False
        self.state = {
            "units": {
                "AUSTRIA": ["A BUD", "F VIE", "F TRI"],
                "ENGLAND": ["F EDI", "F LON", "A LVP"],
                "FRANCE": ["F BRE", "A STP/SC", "A PAR"]
            },
            "centers": {
                "AUSTRIA": ["BUD", "TRI"],
                "ENGLAND": ["EDI", "LON", "TUN"],
                "FRANCE": ["BRE", "STP"]
            },
            "builds": {
                "AUSTRIA": {
                    "count": 0,
                    "homes": []
                },
                "ENGLAND": {
                    "count": 1,
                    "homes": ["TUN"]
                },
                "FRANCE": {
                    "count": -1,
                    "homes": ["PAR"]
                }
            },
            "name": "S1901M"
        }
        self.available_orders = {
            "BUD": ["A BUD H", "A BUD S A EDI - LVP"],
            "VIE": ["A VIE S A LON - TRI", "A VIE S A LON"]
        }

    def get_orderable_locations(self, power):
        orderable_locations = {
            "AUSTRIA": ["BUD", "VIE"]
        }
        return orderable_locations[power]

    def get_all_possible_orders(self):
        return self.available_orders

    def set_orders(self, power, orders):
        pass

    def process(self):
        pass

    def get_state(self):
        return self.state

    def get_phase_history(self):
        return [MockGamePhaseData()]


class MockGamePhaseData:

    def __init__(self):
        self.orders = {
            "AUSTRIA": ["A BUD H", "F VIE - TUN", "F TRI - VIE"],
            "ENGLAND": ["F EDI S F LON", "A LVP S F VIE - TUN"],
            "FRANCE": ["F BRE C A STP/SC - PAR"]
        }
        self.state = MockDiplomacyGameEngine().state
        self.name = "M"


class MockDiplomacy(DiplomacyGame):

    def __init__(self):
        super().__init__([], max_steps=10000, map_name="", callback_list=None)
        self._Diplomacy__game_engine = MockDiplomacyGameEngine()
        self._Diplomacy__players = ["Player1", "Player2", "Player3"]
        self._Diplomacy__players_to_power_names = {"Player1": "AUSTRIA", "Player2": "ENGLAND", "Player3": "FRANCE"}
        self._Diplomacy__power_names_to_players = flip_dictionary(self._Diplomacy__players_to_power_names)


if __name__ == '__main__':
    unittest.main()
