import tensorflow as tf
import tensorflow.keras as keras

from games.dip.dip import get_diplomacy_map_as_adjacency_matrix


class GraphConv(keras.layers.Layer):
    """
    A graph convolution as described in https://arxiv.org/pdf/1609.02907.pdf
    """

    def __init__(self, adjacency_matrix, out_feats, **kwargs):
        """
        :param adjacency_matrix: Adjacency matrix of a graph
        :param out_feats: The number of features in the output after the graph convolution
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.out_feats = out_feats
        adjacency_matrix_with_self_loops = self.add_self_loops(adjacency_matrix)
        self.normalized_adjacency_matrix_with_self_loops = self.normalize_adjacency_matrix(
            adjacency_matrix_with_self_loops)
        self.weight_matrix = None

    def add_self_loops(self, adjacency_matrix):
        return tf.constant(adjacency_matrix + tf.eye(len(adjacency_matrix)))

    def normalize_adjacency_matrix(self, adjacency_matrix):
        degree_matrix = tf.linalg.diag(tf.reduce_sum(adjacency_matrix, axis=0))
        inverse_square_root_of_degree_matrix = tf.linalg.sqrtm(tf.linalg.inv(degree_matrix))
        normalized_adjacency_matrix = inverse_square_root_of_degree_matrix @ adjacency_matrix @ inverse_square_root_of_degree_matrix
        return normalized_adjacency_matrix

    def build(self, input_shape):
        self.weight_matrix = self.add_weight(name="weight_matrix", shape=(input_shape[-1], self.out_feats),
                                             initializer="glorot_normal")
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        return self.normalized_adjacency_matrix_with_self_loops @ inputs @ self.weight_matrix


class GNNLayer(keras.layers.Layer):
    """
    A Graph Neural Network layer as described in https://arxiv.org/pdf/2010.02923.pdf
    """

    def __init__(self, adjacency_matrix, out_feats, use_batch_norm, dropout_probability, **kwargs):
        """
        :param adjacency_matrix: Adjacency matrix of a graph
        :param out_feats: The number of features in the output after the graph convolution
        :param use_batch_norm: Whether to use batch normalization
        :param dropout_probability:
        :param kwargs:
        """
        super(GNNLayer, self).__init__(**kwargs)
        self.out_feats = out_feats
        self.use_batch_norm = use_batch_norm
        self.graph_conv = GraphConv(adjacency_matrix, out_feats=out_feats)
        self.batch_norm = keras.layers.BatchNormalization()
        self.relu = keras.layers.ReLU()
        self.linear = None
        self.drop_out = keras.layers.Dropout(rate=dropout_probability)

    def call(self, features, training=None, **kwargs):
        x = self.graph_conv(features)
        linear_res = features @ self.linear
        x = x + linear_res
        if self.use_batch_norm:
            x = self.batch_norm(x, training=training)
        relu_res = self.relu(x)
        drop_out_res = self.drop_out(relu_res, training=training)
        use_skip_connection = self.in_feats == self.out_feats
        if use_skip_connection:
            return drop_out_res + features
        else:
            return drop_out_res

    def build(self, input_shape):
        self.in_feats = input_shape[-1]
        self.linear = self.add_weight(name="linear_weight_matrix", shape=[input_shape[-1], self.out_feats],
                                      initializer="glorot_normal")
        super().build(input_shape)


class DiplomacyEncoder(keras.layers.Layer):
    """
    Encoder consisting of graph neural network layers.
    The encoder consists of a GNN to encode current state, a GNN
    network to encode the previous move phase, a graph neural network to encode the concatenated
    output from the aforementioned GNNs, an embedding layer to encode previous orders and an embedding
    layer to encode the current season.
    """

    def __init__(self, adjacency_matrix,
                 prev_orders_vocabulary_size,
                 prev_orders_embedding_size,
                 seasons_vocabulary_size,
                 seasons_embedding_size,
                 current_state_encoder_output_size,
                 current_state_encoder_gnn_layer_count,
                 previous_state_encoder_output_size,
                 previous_state_encoder_gnn_layer,
                 concatenated_encoder_output_size,
                 concatenated_encoder_gnn_layer_count,
                 dropout_probability,
                 use_batch_norm,
                 *args, **kwargs):
        """
        :param adjacency_matrix: Adjacency matrix of a graph representation of the diplomacy board
        :param prev_orders_vocabulary_size: The total number of orders in Diplomacy
        :param prev_orders_embedding_size: The size of the embedding of the orders given in the previous move phase
        :param seasons_vocabulary_size: The total number of seasons in Diplomacy
        :param seasons_embedding_size: The size of the embedding of the seasons
        :param current_state_encoder_output_size: The number of output features of the current state encoder
        :param current_state_encoder_gnn_layer_count: The number of GNN layers in the current state encoder
        :param previous_state_encoder_output_size: The number of output features of the previous order encoder
        :param previous_state_encoder_gnn_layer: The number of GNN layers in the previous order encoder
        :param concatenated_encoder_output_size: The number of output features of the concatenated encoder
        :param concatenated_encoder_gnn_layer_count: The number of GNN layers in the concatenated encoder
        :param dropout_probability: Dropout probability in the GNN layers
        :param use_batch_norm: Whether to use batch normalization in the GNN layers
        :param args:
        :param kwargs:
        """
        super(DiplomacyEncoder, self).__init__(*args, **kwargs)
        self.prev_orders_embedding = keras.layers.Embedding(prev_orders_vocabulary_size, prev_orders_embedding_size)
        self.seasons_embedding = keras.layers.Embedding(seasons_vocabulary_size, seasons_embedding_size)
        self.current_state_encoder = [GNNLayer(adjacency_matrix,
                                               current_state_encoder_output_size, use_batch_norm,
                                               dropout_probability)] + [
                                         GNNLayer(adjacency_matrix,
                                                  current_state_encoder_output_size, use_batch_norm,
                                                  dropout_probability) for _ in
                                         range(current_state_encoder_gnn_layer_count - 1)]
        self.prev_state_encoder = [GNNLayer(adjacency_matrix,
                                            previous_state_encoder_output_size, use_batch_norm,
                                            dropout_probability)] + [
                                      GNNLayer(adjacency_matrix,
                                               previous_state_encoder_output_size, use_batch_norm,
                                               dropout_probability) for _ in
                                      range(previous_state_encoder_gnn_layer - 1)]
        self.concatenated_encoder = [
            GNNLayer(adjacency_matrix, concatenated_encoder_output_size,
                     use_batch_norm, dropout_probability) for _ in
            range(concatenated_encoder_gnn_layer_count)]

    def call(self, inputs, training=None, **kwargs):
        """

        :param inputs: state, prev_state, prev_orders, season, build_numbers.
        state: Encoding of the board's current state. shape=[Batch size, #Provinces, Encoding length]
        prev_state: encoding of the boards state in the previous move phase. shape=[Batch size, #Provinces, Encoding length]
        prev_orders: list of orders given in the previous move phase. shape=[Batch size, #Provinces, 1]
        season: the current season. shape=[Batch size, #Provinces, 1]
        build_numbers: current build numbers.shape=[Batch size, #Provinces, 1]
        :param training:
        :param kwargs:
        :return: Encoding of the input
        """
        state, prev_state, prev_orders, season, build_numbers = inputs
        build_numbers = tf.cast(build_numbers, dtype=tf.float32)
        season_embedding = tf.squeeze(self.seasons_embedding(season), axis=2)
        prev_order_embeddings = tf.squeeze(self.prev_orders_embedding(prev_orders), axis=2)
        state_features = tf.concat([state, build_numbers, season_embedding], axis=2)
        prev_state_features = tf.concat([state, prev_order_embeddings, build_numbers, season_embedding], axis=2)
        for cse in self.current_state_encoder:
            state_features = cse(state_features, training=training)
        for pse in self.prev_state_encoder:
            prev_state_features = pse(prev_state_features, training=training)
        concat_features = tf.concat([state_features, prev_state_features], axis=-1)
        for ce in self.concatenated_encoder:
            concat_features = ce(concat_features, training=training)
        return concat_features


class ValueHead(keras.layers.Layer):
    """
    Estimates a state value for each power in Diplomacy
    """

    def __init__(self, power_count, activation, *args, **kwargs):
        """
        :param power_count: The number of powers in the game of Diplomacy
        :param activation: The activation function
        :param args:
        :param kwargs:
        """
        super(ValueHead, self).__init__(*args, **kwargs)
        self.dense = keras.layers.Dense(power_count, activation=activation)

    def call(self, inputs, **kwargs):
        """
        :param inputs: An encoding of the game state (output from the DiplomacyEncoder). shape = [Batch size, #Provinces, #Features]
        :param kwargs:
        :return: A state value estimate for each power
        """
        inputs = tf.reshape(inputs, [tf.shape(inputs)[0]] + [-1])
        return self.dense(inputs)


class PolicyHead(keras.layers.Layer):
    """
    Predicts a policy and chooses actions
    """

    def __init__(self, rnn_type,
                 power_embedding_size,
                 power_embedding_vocabulary_size,
                 predicted_order_embedding_size,
                 predicted_order_vocabulary_size,
                 src_feature_index_range,
                 dst_feature_index_range,
                 **kwargs):
        """
        :param rnn_type: The type of recurrent neural network to be used. Available: 'LSTM' and 'SimpleRNN'
        :param power_embedding_size: The size of the power embeddings
        :param power_embedding_vocabulary_size: The number powers playing in the game of Diplomacy.
        :param predicted_order_embedding_size: THe size of the predicted order embeddings
        :param predicted_order_vocabulary_size: The total number of available orders in Diplomacy, including a default order
        :param src_feature_index_range: The indices where an order's source province is encoded in the available
        order encodings
        :param dst_feature_index_range: The indices where an order's destination province is encoded in the available
         order encodings
        :param kwargs:
        """
        super(PolicyHead, self).__init__(**kwargs)
        self.rnn_type = rnn_type
        self.power_embedding_size = power_embedding_size
        self.predicted_order_embedding_size = predicted_order_embedding_size
        self.predicted_order_vocabulary_size = predicted_order_vocabulary_size
        self.default_order_nummer = self.predicted_order_vocabulary_size - 1
        self.src_feature_index_range = src_feature_index_range
        self.dst_feature_index_range = dst_feature_index_range
        self.power_embedding = keras.layers.Embedding(power_embedding_vocabulary_size, power_embedding_size,
                                                      name="power_embedding")
        self.predicted_order_embedding = keras.layers.Embedding(predicted_order_vocabulary_size,
                                                                predicted_order_embedding_size,
                                                                name="predicted_order_embedding")

    def call(self, inputs, actions=None, training=None, mask=None):
        """
        :param inputs: encoder_output, available_order_encodings, province_owners
        encoder_output: output from the encoder. shape=[Batch size, #Provinces, #Features]
        available_order_encodings: An encoding of each possible order. shape=[Batch size,
        P, #Available orders, #Features]. P is one less than the number of provinces a power needs to win the game.
        Each order in the available_order_encodings has to be encoded as follows: 1. the index of the province
        where the order is given. 2. an integer representation of the order. 3. one-hot encoding
        of the order type. 4. one-hot encoding of the underlying order type. 5. one-hot
        encoding of the source province. 6. one-hot encoding of the destination province.
        7. one-hot encoding of the unit.
        power: The power to select orders for. shape=[Batch size].
        :param actions: Actions that are already chosen. shape=[Batch size, P, 1]
        :param training:
        :param mask:
        :return: chosen orders and their corresponding policy values (probabilities). both with shape=[Batch size, P, 1]
        """
        encoder_output, available_order_encodings, power = inputs
        encoder_output = self.add_row_of_zeros_pr_batch(encoder_output)
        batch_size = self.get_batch_size(encoder_output)
        hidden_state = self.init_lstm_hidden_state(batch_size)
        chosen_order = self.get_default_chosen_order(batch_size)

        power_embeddings = self.power_embedding(power)
        feature_encoding = self.calculate_feature_encoding(encoder_output, available_order_encodings)

        province_indices = self.get_province_indices(available_order_encodings)
        rnn_steps = self.get_province_count(province_indices)
        chosen_orders, chosen_orders_prob = tf.TensorArray(tf.float32, size=rnn_steps), tf.TensorArray(
            tf.float32,
            size=rnn_steps)
        for i in range(rnn_steps):
            province_index = province_indices[:, i]
            province_encoding = self.get_province_encoding(encoder_output, province_index)
            chosen_order_embedding = self.get_chosen_order_embedding(chosen_order)
            rnn_output, hidden_state = self.rnn_iteration(province_encoding, power_embeddings,
                                                          chosen_order_embedding, hidden_state)
            available_order_number_for_province = self.get_available_order_number_for_province(
                available_order_encodings, i)
            feature_encoding_for_province = self.get_feature_encoding_for_province(feature_encoding, i)
            province_logits = self.calculate_logits(feature_encoding_for_province, rnn_output)
            province_logits = self.mask_logits(province_logits, available_order_number_for_province)
            province_probs = self.convert_logits_to_probs(province_logits)
            if actions is not None:
                chosen_order = actions[:, i]
                local_order_idx = self.chosen_order_to_local_order_idx(chosen_order,
                                                                       available_order_number_for_province)
            else:
                local_order_idx = self.predict_order(province_logits)
                chosen_order = self.local_order_idx_to_chosen_order(local_order_idx,
                                                                    available_order_number_for_province)
            chosen_order_prob = self.get_order_prob(province_probs, local_order_idx, chosen_order)
            chosen_orders = chosen_orders.write(i, chosen_order)
            chosen_orders_prob = chosen_orders_prob.write(i, chosen_order_prob)

        return self.permute_output_tensor(chosen_orders.stack()), \
               self.permute_output_tensor(chosen_orders_prob.stack())

    def rnn_iteration(self, province_encoding, power_embeddings, chosen_order_embedding, hidden_state):
        rnn_input = tf.concat([province_encoding, power_embeddings, chosen_order_embedding], axis=1)
        return self.rnn(rnn_input, hidden_state)

    def calculate_feature_encoding(self, encoder_output, available_order_encodings):
        order_src_enc = self.encode_src_feature_for_each_available_order(encoder_output, available_order_encodings)
        order_dst_enc = self.encode_dst_feature_for_each_available_order(encoder_output, available_order_encodings)
        order_enc = self.order_feat_encoder(available_order_encodings[..., 2:])
        return order_src_enc + order_dst_enc + order_enc

    def predict_order(self, logits):
        log_probs = tf.nn.log_softmax(logits)
        return tf.random.categorical(log_probs, num_samples=1)

    def mask_logits(self, logits, available_order_number_for_province):
        mask = available_order_number_for_province != self.default_order_nummer
        return tf.where(mask, logits, tf.float32.min)

    def calculate_logits(self, feature_encoding_for_province, lstm_output):
        lstm_output = tf.expand_dims(lstm_output, axis=1)
        logits = lstm_output @ feature_encoding_for_province
        return tf.squeeze(logits)

    def encode_src_feature_for_each_available_order(self, encoder_output, available_order_encodings):
        order_src_ids = self.find_province_indices(available_order_encodings, self.src_feature_index_range[0],
                                                   self.src_feature_index_range[1])
        src_feature_encoding = self.order_src_feat_encoder(encoder_output)
        return self.map_feature_encoding_to_available_orders(src_feature_encoding, order_src_ids,
                                                             available_order_encodings)

    def encode_dst_feature_for_each_available_order(self, encoder_output, available_order_encodings):
        order_dst_ids = self.find_province_indices(available_order_encodings, self.dst_feature_index_range[0],
                                                   self.dst_feature_index_range[1])
        dst_feature_encoding = self.order_dst_feat_encoder(encoder_output)
        return self.map_feature_encoding_to_available_orders(dst_feature_encoding, order_dst_ids,
                                                             available_order_encodings)

    def map_feature_encoding_to_available_orders(self, feature_encoding, order_ids, available_order_encodings):
        updates = tf.gather_nd(feature_encoding, tf.gather(order_ids, [0, 3], axis=1))
        return tf.scatter_nd(order_ids[:, :-1], updates,
                             shape=tf.concat(
                                 [tf.shape(available_order_encodings, out_type=tf.int64)[:-1], [self.rnn_units]],
                                 axis=0))

    def local_order_idx_to_chosen_order(self, local_order_idx, available_order_number_for_province):
        return tf.gather(available_order_number_for_province, local_order_idx, axis=-1,
                         batch_dims=1)

    def chosen_order_to_local_order_idx(self, global_idx, available_order_number_for_province):
        batch_size = self.get_batch_size(available_order_number_for_province)
        return tf.stack([tf.where(global_idx[i] == available_order_number_for_province[i])[0] for
                         i in range(batch_size)])

    def get_chosen_order_embedding(self, chosen_order):
        return tf.squeeze(self.predicted_order_embedding(chosen_order), axis=1)

    def get_order_prob(self, probs, idx, chosen_order):
        order_prob = tf.gather(probs, idx, batch_dims=1, axis=-1)
        return tf.where(chosen_order == self.default_order_nummer, 1.0, order_prob)

    def get_province_encoding(self, encoder_output, province_index):
        return tf.gather(encoder_output, province_index, axis=1, batch_dims=1)

    def find_province_indices(self, encoding, start_index, end_index):
        return tf.where(encoding[..., start_index:end_index] == 1)

    def convert_logits_to_probs(self, logits):
        return tf.nn.softmax(logits)

    def get_province_count(self, province_indices):
        return province_indices.get_shape()[-1]

    def get_province_indices(self, available_order_encodings):
        return tf.cast(available_order_encodings[:, :, 0, 0], dtype=tf.int32)

    def get_available_order_number_for_province(self, available_order_encodings, i):
        return available_order_encodings[:, i, :, 1]

    def get_feature_encoding_for_province(self, feature_encoding, i):
        return tf.transpose(feature_encoding[:, i], perm=[0, 2, 1])

    def get_default_chosen_order(self, batch_size):
        return tf.fill((batch_size, 1), tf.cast(self.default_order_nummer, dtype=tf.float32))

    def init_lstm_hidden_state(self, batch_size):
        return [tf.zeros([batch_size, self.rnn_units]), tf.zeros([batch_size, self.rnn_units])]

    def permute_output_tensor(self, tensor):
        return tf.transpose(tensor, perm=[1, 0, 2])

    def add_row_of_zeros_pr_batch(self, tensor):
        return tf.pad(tensor, [[0, 0], [0, 1], [0, 0]])

    def get_batch_size(self, tensor):
        return tf.shape(tensor)[0]

    def build(self, input_shape):
        self.rnn_units = input_shape[0][-1] + self.power_embedding_size + self.predicted_order_embedding_size
        self.order_src_feat_encoder = keras.layers.Dense(self.rnn_units, use_bias=False,
                                                         name="order_src_feat_encoder")
        self.order_dst_feat_encoder = keras.layers.Dense(self.rnn_units, use_bias=False,
                                                         name="order_dst_feat_encoder")
        self.order_feat_encoder = keras.layers.Dense(self.rnn_units, name="order_feat_encoder")
        self.rnn = get_recurrent_network(self.rnn_type, self.rnn_units)
        super().build(input_shape)


class DipValueNetwork(keras.Model):
    """
    Value network used to calculate state values for Diplomacy game states
    """

    def __init__(self, encoder, power_count, activation, optimizer, learning_rate, loss_fn, *args, **kwargs):
        """
        :param encoder: An encoder which encodes game states 
        :param power_count: The number of powers in the Diplomacy game
        :param activation: The activation function of the value head
        :param optimizer: One of "adam", "adagrad", "rmsprop" and "sgd"
        :param learning_rate: 
        :param loss_fn:
        :param args: 
        :param kwargs: 
        """
        super().__init__(*args, **kwargs)
        self.encoder = encoder
        self.value_head = ValueHead(power_count, activation, name="value_head")
        self.value_function_loss = keras.losses.get(loss_fn)
        self.optimizer = get_optimizer(optimizer, learning_rate)

    def call(self, inputs, training=None, mask=None):
        """
        :param inputs: state, prev_state, prev_orders, season, build_numbers.
        state: Encoding of the board's current state. shape=[Batch size, #Provinces, Encoding length]
        prev_state: encoding of the boards state in the previous move phase. shape=[Batch size, #Provinces, Encoding length]
        prev_orders: list of orders given in the previous move phase. shape=[Batch size, #Provinces, 1]
        season: the current season. shape=[Batch size, #Provinces, 1]
        build_numbers: current build numbers.shape=[Batch size, #Provinces, 1]
        :param training:
        :param mask:
        :return:
        """
        state, prev_state, prev_orders, season, build_numbers = inputs
        state_enc = self.encoder([state, prev_state, prev_orders, season, build_numbers], training=training)
        return self.value_head(state_enc)

    def get_config(self):
        base_config = super().get_config()
        return {**base_config}


class DipPolicyNetwork(keras.Model):
    """
    Policy network used to predict orders and their corresponding probabilities
    """

    def __init__(self, encoder, optimizer, learning_rate, policy_head, *args, **kwargs):
        """
        :param encoder: An encoder which encodes game states
        :param optimizer: One of "adam", "adagrad", "rmsprop" and "sgd"
        :param learning_rate:
        :param policy_head: PolicyHead used to predict orders and their corresponding probabilities
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.encoder = encoder
        self.policy_head = policy_head
        self.optimizer = get_optimizer(optimizer, learning_rate)

    def call(self, inputs, actions=None, training=None, mask=None):
        """
        :param inputs: encoder_input (see the DiplomacyEncoder documentation for a more thorough description) and policy_head_input.
        policy_head_input consists of [available_order_encodings, power] (see the PolicyHead documentation
        for a more thorough
        description)
        :param actions: Actions that are already chosen. shape=[Batch size, P, 1]
        :param training:
        :param mask:
        :return:
        """
        encoder_input, policy_head_input = inputs
        state_enc = self.encoder(encoder_input, training=training)
        return self.policy_head([state_enc, policy_head_input[0], policy_head_input[1]], actions=actions)

    def get_config(self):
        base_config = super().get_config()
        return {**base_config}


def get_optimizer(name, learning_rate):
    if name == "adam":
        return keras.optimizers.Adam(learning_rate=learning_rate)
    elif name == "adagrad":
        return keras.optimizers.Adagrad(learning_rate=learning_rate)
    elif name == "rmsprop":
        return keras.optimizers.RMSprop(learning_rate=learning_rate)
    elif name == "sgd":
        return keras.optimizers.SGD(learning_rate=learning_rate)


def get_recurrent_network(name, units):
    if name == "LSTM":
        return keras.layers.LSTMCell(units, name=name)
    elif name == "SimpleRNN":
        return keras.layers.SimpleRNNCell(units, name=name)


class NeuralNetworkFactory:

    @staticmethod
    def create_neural_networks(game, neural_network_parameters, **kwargs):
        if game == "Diplomacy":
            adjacency_matrix = get_diplomacy_map_as_adjacency_matrix(map_name=kwargs["map_name"])
            return NeuralNetworkFactory.get_diplomacy_networks(adjacency_matrix, neural_network_parameters)
        raise ValueError(f"Cannot find neural networks suitable to learn {game}")

    @staticmethod
    def get_diplomacy_networks(adjacency_matrix, neural_network_parameters):
        encoder = DiplomacyEncoder(adjacency_matrix=adjacency_matrix,
                                   prev_orders_vocabulary_size=neural_network_parameters["encoder"][
                                       "prev_orders_vocabulary_size"],
                                   prev_orders_embedding_size=neural_network_parameters["encoder"][
                                       "prev_orders_embedding_size"],
                                   seasons_vocabulary_size=neural_network_parameters["encoder"][
                                       "seasons_vocabulary_size"],
                                   seasons_embedding_size=neural_network_parameters["encoder"][
                                       "seasons_embedding_size"],
                                   current_state_encoder_output_size=neural_network_parameters["encoder"][
                                       "current_state_encoder_output_size"],
                                   current_state_encoder_gnn_layer_count=neural_network_parameters["encoder"][
                                       "current_state_encoder_gcn_depth"],
                                   previous_state_encoder_output_size=neural_network_parameters["encoder"][
                                       "previous_state_encoder_output_size"],
                                   previous_state_encoder_gnn_layer=neural_network_parameters["encoder"][
                                       "previous_state_encoder_gcn_depth"],
                                   concatenated_encoder_output_size=neural_network_parameters["encoder"][
                                       "concatenated_encoder_output_size"],
                                   concatenated_encoder_gnn_layer_count=neural_network_parameters["encoder"][
                                       "concatenated_encoder_gcn_depth"],
                                   dropout_probability=neural_network_parameters["encoder"]["dropout_probability"],
                                   use_batch_norm=neural_network_parameters["encoder"]["use_batch_norm"])
        value_net = DipValueNetwork(encoder=encoder,
                                    power_count=neural_network_parameters["value"]["output_size"],
                                    activation=neural_network_parameters["value"]["activation"],
                                    optimizer=neural_network_parameters["value"]["optimizer"],
                                    learning_rate=neural_network_parameters["value"]["learning_rate"],
                                    loss_fn=neural_network_parameters["value"]["loss"])

        policy_head = PolicyHead(power_embedding_size=neural_network_parameters["policy"]["power_embedding_size"],
                                 power_embedding_vocabulary_size=neural_network_parameters["policy"][
                                     "power_embedding_vocabulary_size"],
                                 predicted_order_embedding_size=neural_network_parameters["policy"][
                                     "predicted_order_embedding_size"],
                                 predicted_order_vocabulary_size=neural_network_parameters["policy"][
                                     "predicted_order_vocabulary_size"],
                                 src_feature_index_range=neural_network_parameters["policy"]["src_feature_index_range"],
                                 dst_feature_index_range=neural_network_parameters["policy"]["dst_feature_index_range"],
                                 name="policy_head",
                                 rnn_type=neural_network_parameters["policy"]["recurrent_network_type"])
        policy_net = DipPolicyNetwork(encoder=encoder,
                                      optimizer=neural_network_parameters["policy"]["optimizer"],
                                      learning_rate=neural_network_parameters["policy"]["learning_rate"],
                                      policy_head=policy_head)
        return policy_net, value_net
