import shutil
import unittest
import os

os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
import tensorflow as tf
import numpy as np
import keras.models
from neural_networks.networks import GNNLayer, DiplomacyEncoder, ValueHead, PolicyHead, GraphConv, DipValueNetwork, \
    DipPolicyNetwork
from utils.utils import create_random_integer_symmetric_matrix, random_one_hot

np.random.seed(42)
tf.random.set_seed(42)

class TestGraphConv(unittest.TestCase):

    def setUp(self):
        self.graph_conv = GraphConv(np.array([[0, 1, 1], [1, 0, 0], [1, 0, 0]]), out_feats=7)
        self.expected_normalized_adjacency_matrix = np.sqrt(np.array([[1 / 3 * 1 / 3, 1 / 3 * 1 / 2, 1 / 3 * 1 / 2],
                                                                      [1 / 3 * 1 / 2, 1 / 2 * 1 / 2, 0],
                                                                      [1 / 3 * 1 / 2, 0, 1 / 2 * 1 / 2]]))

    def test_if_correct_adjacency_matrix_is_created(self):
        np.testing.assert_almost_equal(self.expected_normalized_adjacency_matrix,
                                       self.graph_conv.normalized_adjacency_matrix_with_self_loops)

    def test_output_shape(self):
        res = self.graph_conv(np.random.uniform(0, 1, size=[3, 5]))
        self.assertEqual((3, 7), res.shape)

    def test_if_gradient_is_calculated(self):
        with tf.GradientTape() as tape:
            res = self.graph_conv(np.random.uniform(0, 1, size=[3, 5]))
            loss = (res - tf.random.uniform(shape=[3, 7]))
        gradient = tape.gradient(loss, self.graph_conv.trainable_weights)
        self.assertEqual(1, len(gradient))
        self.assertEqual((5, 7), gradient[0].shape)

    def test_batch_input(self):
        res = self.graph_conv(np.random.uniform(0, 1, size=[2, 3, 5]))
        self.assertEqual((2, 3, 7), res.shape)


class TestGNNLayer(unittest.TestCase):

    def setUp(self):
        self.provinces = 10
        self.out_feats = 120
        self.adjacency_matrix = create_random_integer_symmetric_matrix(0, 2, (self.provinces, self.provinces))
        self.gnn_layer = GNNLayer(self.adjacency_matrix, out_feats=self.out_feats,
                                  dropout_probability=0.4, use_batch_norm=True)

    def test_gnn_layer_output_shape(self):
        expected_output_shape = (2, self.provinces, self.out_feats)
        res = self.apply_gnn_layer(self.gnn_layer, 35, 2)
        self.assertEqual(expected_output_shape, res.shape)

    def apply_gnn_layer(self, gnn_layer, in_feats, batch_size):
        input_shape = (batch_size, self.provinces, in_feats)
        gnn_input = np.random.random(size=input_shape)
        return gnn_layer(gnn_input).numpy()


class TestEncoder(unittest.TestCase):

    def setUp(self):
        self.batch_size = 2
        self.provinces = 10
        self.prev_orders_vocabulary_size = 10
        self.season_vocabulary_size = 10
        self.concatenated_encoder_output_size = 240
        self.current_state_input_shape = (self.batch_size, self.provinces, 35)
        self.prev_orders_input_shape = (self.batch_size, self.provinces, 35)
        self.adjacency_matrix = create_random_integer_symmetric_matrix(0, 2, (
            self.current_state_input_shape[-2], self.current_state_input_shape[-2]))
        self.encoder = DiplomacyEncoder(adjacency_matrix=self.adjacency_matrix,
                                        prev_orders_vocabulary_size=self.prev_orders_vocabulary_size,
                                        prev_orders_embedding_size=10,
                                        seasons_vocabulary_size=self.season_vocabulary_size,
                                        seasons_embedding_size=3,
                                        current_state_encoder_output_size=120,
                                        current_state_encoder_gnn_layer_count=2,
                                        previous_state_encoder_output_size=120,
                                        previous_state_encoder_gnn_layer=2,
                                        concatenated_encoder_output_size=self.concatenated_encoder_output_size,
                                        concatenated_encoder_gnn_layer_count=2,
                                        dropout_probability=0.2,
                                        use_batch_norm=True)

    def test_encoder_output_shape(self):
        current_state_input = np.random.random(self.current_state_input_shape)
        previous_state_input = np.random.random(self.prev_orders_input_shape)
        prev_orders = np.random.randint(0, self.prev_orders_vocabulary_size,
                                        size=(self.batch_size, self.provinces, 1))
        seasons = np.random.randint(0, self.season_vocabulary_size, size=(self.batch_size, self.provinces, 1))
        build_numbers = np.random.randint(-5, 5, size=(self.batch_size, self.provinces, 1))
        res = self.encoder([current_state_input, previous_state_input, prev_orders,
                            seasons, build_numbers])
        self.assertEqual((self.batch_size, self.provinces, self.concatenated_encoder_output_size), res.shape)

    def test_season_not_in_vocabulary(self):
        current_state_input = np.random.random(self.current_state_input_shape)
        previous_state_input = np.random.random(self.prev_orders_input_shape)
        prev_orders = np.random.randint(0, self.prev_orders_vocabulary_size,
                                        size=(self.batch_size, self.provinces, 1))
        seasons = np.full(fill_value=self.season_vocabulary_size, shape=(self.batch_size, self.provinces, 1))
        build_numbers = np.random.randint(-5, 5, size=(self.batch_size, self.provinces, 1))
        self.assertRaises(tf.errors.InvalidArgumentError, self.encoder,
                          [current_state_input, previous_state_input, prev_orders,
                           seasons, build_numbers])

    def test_prev_orders_not_in_vocabulary(self):
        current_state_input = np.random.random(self.current_state_input_shape)
        previous_state_input = np.random.random(self.prev_orders_input_shape)
        prev_orders = np.full(fill_value=self.prev_orders_vocabulary_size,
                              shape=(self.batch_size, self.provinces, 1))
        seasons = np.random.randint(0, self.season_vocabulary_size, size=(self.batch_size, self.provinces, 1))
        build_numbers = np.random.randint(-5, 5, size=(self.batch_size, self.provinces, 1))
        self.assertRaises(tf.errors.InvalidArgumentError, self.encoder,
                          [current_state_input, previous_state_input, prev_orders,
                           seasons, build_numbers])


class TestValueHead(unittest.TestCase):

    def setUp(self):
        self.value_head = ValueHead(1, "relu")

    def test_value_head_output_shape(self):
        res = self.value_head(tf.random.uniform(shape=[1, 19440]))
        self.assertEqual((1, 1), res.shape)

        res = self.value_head(tf.random.uniform(shape=[1, 81, 240]))
        self.assertEqual((1, 1), res.shape)

    def test_value_head_batch_output_shape(self):
        res = self.value_head(tf.random.uniform(shape=[10, 19440]))
        self.assertEqual((10, 1), res.shape)

        res = self.value_head(tf.random.uniform(shape=[10, 81, 240]))
        self.assertEqual((10, 1), res.shape)


class TestPolicyHead(unittest.TestCase):

    def setUp(self):
        self.provinces = 10
        self.encoder_output_features = 100
        self.batch_size = 3
        self.predicted_order_vocabulary_size = 10
        self.available_order_count = 10
        self.order_type_count = 3
        self.unit_count = 3
        self.powers = 5
        self.rnn_steps = 5
        self.src_feature_index_range = [1 + self.order_type_count + self.order_type_count,
                                        1 + self.order_type_count + self.order_type_count + self.provinces]
        self.dst_feature_index_range = [1 + self.order_type_count + self.order_type_count + self.provinces,
                                        1 + self.order_type_count + self.order_type_count + 2 * self.provinces]
        self.available_order_encoding = get_mock_available_order_encoding(
            available_order_count=self.available_order_count,
            batch_size=self.batch_size,
            order_vocabulary_size=self.predicted_order_vocabulary_size,
            rnn_steps=self.rnn_steps,
            province_count=self.provinces,
            order_type_count=self.order_type_count,
            unit_count=self.unit_count)

        self.ph = PolicyHead(rnn_type="LSTM",
                             power_embedding_size=60,
                             power_embedding_vocabulary_size=self.powers,
                             predicted_order_embedding_size=40,
                             predicted_order_vocabulary_size=self.predicted_order_vocabulary_size + 1,
                             # pluss one due to default order
                             src_feature_index_range=self.src_feature_index_range,
                             dst_feature_index_range=self.dst_feature_index_range)

    def test_policy_head_correct_output_shape(self):
        encoder_output = np.random.random(size=(self.batch_size, self.provinces, self.encoder_output_features))
        power = np.random.randint(0, self.powers, size=(self.batch_size,))
        chosen_orders, chosen_orders_probs = self.ph([encoder_output, self.available_order_encoding, power])
        self.assertEqual((self.batch_size, self.rnn_steps, 1), chosen_orders.shape)
        self.assertEqual((self.batch_size, self.rnn_steps, 1), chosen_orders_probs.shape)

    def test_outputs_correct_actions_when_actions_is_provided(self):
        encoder_output = np.random.random(size=(self.batch_size, self.provinces, self.encoder_output_features))
        power = np.random.randint(0, self.powers, size=(self.batch_size,))
        chosen_orders, chosen_orders_probs = self.ph([encoder_output, self.available_order_encoding, power])
        actions = chosen_orders
        chosen_orders, chosen_orders_probs = self.ph([encoder_output, self.available_order_encoding, power],
                                                     actions=actions)
        np.testing.assert_equal(actions.numpy(), chosen_orders.numpy())

    def test_gradient_calculation(self):
        encoder_output = np.random.random(size=(self.batch_size, self.provinces, self.encoder_output_features))
        power = np.random.randint(0, self.powers, size=(self.batch_size,))
        with tf.GradientTape() as tape:
            chosen_orders, chosen_orders_probs = self.ph([encoder_output, self.available_order_encoding, power])
            loss = ((tf.stop_gradient(chosen_orders_probs + 1)) - chosen_orders_probs) ** 2
        gradient = tape.gradient(loss, self.ph.trainable_weights)
        for g in gradient:
            if isinstance(g, tf.IndexedSlices):
                self.assertFalse(np.isnan(g.values).any())
            else:
                self.assertFalse(np.isnan(g).any())
        self.assertEqual(len(self.ph.trainable_weights), len(gradient))


class TestDipValueNetwork(unittest.TestCase):

    def setUp(self):
        self.powers = 5
        self.batch_size = 2
        self.provinces = 10
        self.prev_orders_vocabulary_size = 10
        self.season_vocabulary_size = 10
        self.concatenated_encoder_output_size = 240
        self.current_state_input_shape = (self.batch_size, self.provinces, 35)
        self.prev_orders_input_shape = (self.batch_size, self.provinces, 35)
        self.adjacency_matrix = create_random_integer_symmetric_matrix(0, 2, (
            self.current_state_input_shape[-2], self.current_state_input_shape[-2]))
        self.encoder = DiplomacyEncoder(adjacency_matrix=self.adjacency_matrix,
                                        prev_orders_vocabulary_size=self.prev_orders_vocabulary_size,
                                        prev_orders_embedding_size=10,
                                        seasons_vocabulary_size=self.season_vocabulary_size,
                                        seasons_embedding_size=3,
                                        current_state_encoder_output_size=120,
                                        current_state_encoder_gnn_layer_count=2,
                                        previous_state_encoder_output_size=120,
                                        previous_state_encoder_gnn_layer=2,
                                        concatenated_encoder_output_size=self.concatenated_encoder_output_size,
                                        concatenated_encoder_gnn_layer_count=2,
                                        dropout_probability=0.2,
                                        use_batch_norm=True)
        self.value_network = DipValueNetwork(encoder=self.encoder,
                                             power_count=self.powers,
                                             activation="relu",
                                             optimizer="adam",
                                             learning_rate=0.05,
                                             loss_fn="mse")

    def test_correct_output_shape(self):
        state_value = self.calculate_state_values(self.value_network)
        self.assertEqual((self.batch_size, self.powers), state_value.shape)

    def test_can_be_saved_and_restored(self):
        save_path = "saved_models_test"
        self.calculate_state_values(self.value_network)
        self.value_network.save(save_path)
        value_network = keras.models.load_model(save_path)
        state_value = self.calculate_state_values(value_network)
        self.assertEqual((self.batch_size, self.powers), state_value.shape)
        shutil.rmtree(save_path)

    def calculate_state_values(self, value_network):
        current_state_input = np.random.random(self.current_state_input_shape)
        previous_state_input = np.random.random(self.prev_orders_input_shape)
        prev_orders = np.random.randint(0, self.prev_orders_vocabulary_size,
                                        size=(self.batch_size, self.provinces, 1))
        seasons = np.random.randint(0, self.season_vocabulary_size, size=(self.batch_size, self.provinces, 1))
        build_numbers = np.random.randint(-5, 5, size=(self.batch_size, self.provinces, 1))
        return value_network([current_state_input, previous_state_input, prev_orders, seasons,
                              build_numbers])


class TestDipPolicyNetwork(unittest.TestCase):

    def setUp(self):
        self.powers = 5
        self.batch_size = 2
        self.provinces = 10
        self.prev_orders_vocabulary_size = 10
        self.season_vocabulary_size = 10
        self.concatenated_encoder_output_size = 240
        self.current_state_input_shape = (self.batch_size, self.provinces, 35)
        self.prev_orders_input_shape = (self.batch_size, self.provinces, 35)
        self.adjacency_matrix = create_random_integer_symmetric_matrix(0, 2, (
            self.current_state_input_shape[-2], self.current_state_input_shape[-2]))
        self.available_order_count = 10
        self.rnn_steps = 5
        self.unit_count = 3
        self.order_type_count = 3
        self.predicted_order_vocabulary_size = 10
        self.order_type_count = 3
        self.predicted_order_vocabulary_size = 10
        self.src_feature_index_range = [1 + self.order_type_count + self.order_type_count,
                                        1 + self.order_type_count + self.order_type_count + self.provinces]
        self.dst_feature_index_range = [1 + self.order_type_count + self.order_type_count + self.provinces,
                                        1 + self.order_type_count + self.order_type_count + 2 * self.provinces]
        self.available_order_encoding = get_mock_available_order_encoding(
            available_order_count=self.available_order_count,
            batch_size=self.batch_size,
            order_vocabulary_size=self.predicted_order_vocabulary_size,
            rnn_steps=self.rnn_steps,
            province_count=self.provinces,
            order_type_count=self.order_type_count,
            unit_count=self.unit_count)
        self.encoder = DiplomacyEncoder(adjacency_matrix=self.adjacency_matrix,
                                        prev_orders_vocabulary_size=self.prev_orders_vocabulary_size,
                                        prev_orders_embedding_size=10,
                                        seasons_vocabulary_size=self.season_vocabulary_size,
                                        seasons_embedding_size=3,
                                        current_state_encoder_output_size=120,
                                        current_state_encoder_gnn_layer_count=2,
                                        previous_state_encoder_output_size=120,
                                        previous_state_encoder_gnn_layer=2,
                                        concatenated_encoder_output_size=self.concatenated_encoder_output_size,
                                        concatenated_encoder_gnn_layer_count=2,
                                        dropout_probability=0.2,
                                        use_batch_norm=True)
        self.policy_head = PolicyHead(rnn_type="LSTM",
                                      power_embedding_size=60,
                                      power_embedding_vocabulary_size=self.powers,
                                      predicted_order_embedding_size=40,
                                      predicted_order_vocabulary_size=self.predicted_order_vocabulary_size + 1,
                                      # pluss one due to default order
                                      src_feature_index_range=self.src_feature_index_range,
                                      dst_feature_index_range=self.dst_feature_index_range)

        self.policy_network = DipPolicyNetwork(encoder=self.encoder,
                                               optimizer="adam",
                                               learning_rate=0.05,
                                               policy_head=self.policy_head)

    def test_correct_output_shape(self):
        chosen_orders, chosen_orders_probs = self.calculate_policy(self.policy_network)
        self.assertEqual((self.batch_size, self.rnn_steps, 1), chosen_orders.shape)
        self.assertEqual((self.batch_size, self.rnn_steps, 1), chosen_orders_probs.shape)

    def test_can_be_saved_and_restored(self):
        save_path = "saved_models_test"
        self.calculate_policy(self.policy_network)
        self.policy_network.save(save_path)
        value_network = keras.models.load_model(save_path)
        chosen_orders, chosen_orders_probs = self.calculate_policy(value_network)
        self.assertEqual((self.batch_size, self.rnn_steps, 1), chosen_orders.shape)
        self.assertEqual((self.batch_size, self.rnn_steps, 1), chosen_orders_probs.shape)
        shutil.rmtree(save_path)

    def calculate_policy(self, policy_network):
        current_state_input = np.random.random(self.current_state_input_shape)
        previous_state_input = np.random.random(self.prev_orders_input_shape)
        prev_orders = np.random.randint(0, self.prev_orders_vocabulary_size,
                                        size=(self.batch_size, self.provinces, 1))
        seasons = np.random.randint(0, self.season_vocabulary_size, size=(self.batch_size, self.provinces, 1))
        build_numbers = np.random.randint(-5, 5, size=(self.batch_size, self.provinces, 1))
        encoder_input = [current_state_input, previous_state_input, prev_orders, seasons,
                         build_numbers]
        policy_head_input = [self.available_order_encoding, np.random.randint(0, self.powers, size=(self.batch_size,))]
        return policy_network([encoder_input, policy_head_input])


def get_mock_available_order_encoding(available_order_count,
                                      batch_size,
                                      order_vocabulary_size,
                                      province_count,
                                      rnn_steps,
                                      order_type_count,
                                      unit_count):
    order_index = np.random.randint(0, order_vocabulary_size, size=(batch_size, rnn_steps, available_order_count, 1))
    location_index = np.random.randint(0, province_count, size=(batch_size, rnn_steps, available_order_count, 1))
    order_type = random_one_hot(order_type_count, (batch_size, rnn_steps, available_order_count))
    unit = random_one_hot(unit_count, (batch_size, rnn_steps, available_order_count))
    underlying_order_type = random_one_hot(order_type_count, (batch_size, rnn_steps, available_order_count))
    src_location = random_one_hot(province_count, (batch_size, rnn_steps, available_order_count))
    dst_location = random_one_hot(province_count, (batch_size, rnn_steps, available_order_count))
    return np.concatenate(
        (location_index, order_index, order_type, underlying_order_type, src_location, dst_location,
         unit), axis=-1)


if __name__ == '__main__':
    unittest.main()
