import pandas as pd
from diplomacy import Game
from ast import literal_eval
from copy import deepcopy
import re
import os
import matplotlib.pyplot as plt
import seaborn as sns

from utils.utils import print_walk_error


def calculate_and_print_log_stats(dir_path, map_name, include_x_power_support_stats):
    for root, dirs, files in os.walk(dir_path, onerror=print_walk_error):
        for filename in files:
            if filename.endswith("tournament_log.csv"):
                path = root + "/" + filename
                print("\nStatistics for", path, ":")
                calculate_and_print_log_stats_for_single_file(path, map_name, include_x_power_support_stats)


def calculate_and_print_log_stats_for_single_file(log_file_path, map_name, include_x_power_support_stats):
    df = pd.read_csv(log_file_path, sep=";")
    print_game_count(df)
    print_step_count(df)
    calculate_and_print_avg_rewards(df)
    print_win_percentage(df)
    print_advancement_info(df)
    print_attack_count(df)
    print_support_count(df)
    print_hold_count(df)
    if include_x_power_support_stats:
        calculate_and_print_cross_power_support(df, map_name)


def calculate_and_plot_training_log_stats(dir_path, players_to_plot_stats_for="player1"):
    training_results = pd.DataFrame(columns=["player_name", "tournament", "episode", "avg_reward"])
    for root, dirs, files in os.walk(dir_path, onerror=print_walk_error):
        if len(files) > 0:
            training_results = aggregate_training_results(training_results, root, files, players_to_plot_stats_for)
    training_results = training_results.reset_index()
    sns.relplot(data=training_results, x="episode", y="avg_reward", col="player_name", row="tournament",
                kind="line", facet_kws={"sharey": "row"}, ci="sd")
    plt.show()


def aggregate_training_results(training_results, root, files, player_to_plot_stats_for="player1"):
    for filename in files:
        match = re.match(f"(?P<tournament_name>.*)-episode-(?P<episode>[0-9]*)-test_log.csv", filename)
        if match:
            file_path = root + "/" + filename
            df = pd.read_csv(file_path, sep=";")
            avg_rewards = calculate_avg_rewards(df)
            avg_rewards = avg_rewards.loc[avg_rewards["player_name"] == player_to_plot_stats_for]
            algorithm = dir_path_to_algorithm(root)
            if algorithm:
                avg_rewards["player_name"] = algorithm
            avg_rewards["episode"] = int(match.group("episode"))
            avg_rewards["tournament"] = match.group("tournament_name")
            training_results = pd.concat([training_results, avg_rewards])
    return training_results


def dir_path_to_algorithm(dir_path):
    if "A2C" in dir_path.upper():
        return "A2C"
    elif "LOLA" in dir_path.upper():
        return "LOLA"


def print_step_count(df):
    power_count = get_power_count(df)
    print(f"#steps: {df.shape[0] / power_count}")


def print_game_count(df):
    print(f"#Games: {get_game_count(df)}")


def calculate_and_print_avg_rewards(df):
    rewards = calculate_avg_rewards(df)
    for index, row in rewards.iterrows():
        print(f"""Avg. reward {row["player_name"]}: {row["avg_reward"]} , STD: {row["std_reward"]}""")


def calculate_avg_rewards(df):
    """
    Calculate avg rewards. Important: assumes that non-zero rewards are only given at a game's last step
    :param df:
    :return:
    """
    game_count = get_game_count(df)
    last_steps = get_last_step_from_each_game(df)
    num_powers = len(last_steps["power"].unique())
    game_numbers = []
    for i in range(game_count):
        for _ in range(num_powers):
            game_numbers.append(i)
    last_steps.insert(0, "game_number", game_numbers)
    last_steps_agg = last_steps.groupby(["player_name", "game_number"])["reward"].sum().reset_index()
    # adapted from https://stackoverflow.com/questions/26599347/groupby-pandas-dataframe-and-calculate-mean-and-stdev-of-one-column-and-add-the
    last_steps_agg = last_steps_agg.groupby(["player_name"], as_index=False).agg({"reward": ["mean", "std"]})
    # adapted from https://stackoverflow.com/questions/14507794/pandas-how-to-flatten-a-hierarchical-index-in-columns
    last_steps_agg.columns = ["_".join(a) for a in last_steps_agg.columns.to_flat_index()]
    return last_steps_agg.rename(columns={"player_name_": "player_name", "reward_mean": "avg_reward",
                                          "reward_std": "std_reward"})


def print_win_percentage(df):
    game_count = get_game_count(df)
    player_names = df["player_name"].unique()
    win_percentages = [get_win_count(df, player_name) * 100 / game_count for player_name in player_names]
    for i in range(len(win_percentages)):
        print(f"Win percentage {player_names[i]}: {win_percentages[i]}%")


def print_advancement_info(df):
    if df[df["orders"].str.contains(",")].count()["orders"] == 0:
        print("Log file doesn't contain an advancement")

    else:
        print("Log file contains an advancement from a player")


def print_attack_count(df):
    player_names = get_player_names(df)
    attack_counts = [get_attack_count(df, player_name) for player_name in player_names]
    for i in range(len(attack_counts)):
        print(f"Attack counts {player_names[i]}: {attack_counts[i]}")


def print_support_count(df):
    player_names = get_player_names(df)
    support_counts = [get_support_count(df, player_name) for player_name in player_names]
    for i in range(len(support_counts)):
        print(f"Support counts {player_names[i]}: {support_counts[i]}")


def print_hold_count(df):
    player_names = get_player_names(df)
    hold_counts = [get_hold_count(df, player_name) for player_name in player_names]
    for i in range(len(hold_counts)):
        print(f"Hold counts {player_names[i]}: {hold_counts[i]}")


def get_attack_count(df, player_name):
    orders = get_players_orders(df, player_name)
    return orders.apply(count_attacks).sum()


def get_support_count(df, player_name):
    orders = get_players_orders(df, player_name)
    return orders.apply(count_supports).sum()


def get_hold_count(df, player_name):
    orders = get_players_orders(df, player_name)
    return orders.apply(count_holds).sum()


def get_players_orders(df, player_name):
    player_records = df[df["player_name"] == player_name]
    return player_records["orders"].apply(eval)


def count_attacks(li):
    return sum_by_regex(li, "^[AF][\u0020][A-Z]{3}[\u0020][-][\u0020][A-Z]{3}")


def count_supports(li):
    return sum_by_regex(li, "^[AF][\u0020][A-Z]{3}[\u0020][S]")


def count_holds(li):
    return sum_by_regex(li, "^[AF][\u0020][A-Z]{3}[\u0020][H]")


def sum_by_regex(li, regex):
    return sum(re.match(regex, e) is not None for e in li)


def get_player_names(df):
    return df["player_name"].unique()


def get_game_count(df):
    df["step_shift"] = df["step"].shift(1)
    return df[df["step_shift"] > df["step"]].count()["step"] + 1


def get_win_count(df, player_name):
    return df.loc[(df["reward"] == 1) & (df["player_name"] == player_name)].shape[0]


def get_power_count(df):
    return df["power"].unique().shape[0]


def get_last_step_from_each_game(df):
    power_count = get_power_count(df)
    df["step_shift"] = df["step"].shift(-power_count)
    df["step_shift"].fillna(1, inplace=True)
    return df[df["step_shift"] == 1]


def calculate_and_print_cross_power_support(df, map_name):
    game = Game(map_name=map_name)
    prev_step = 1
    power_to_orders = {}
    x_support_stats = []
    for index, row in df.iterrows():
        step, power, orders = row["step"], row["power"], literal_eval(row["orders"])
        if step != prev_step:
            set_orders(power_to_orders, game)
            x_support_stats.append(compute_xpower_supports(power_to_orders, game))
            game.process()
        if step < prev_step:
            game = Game(map_name=map_name)
        power_to_orders[power] = orders
        prev_step = step
        game.set_orders(power, orders)
    set_orders(power_to_orders, game)
    x_support_stats.append(compute_xpower_supports(power_to_orders, game))
    support_count = sum([xss.num_supports for xss in x_support_stats])
    cross_power_support_count = sum([xss.num_xpower for xss in x_support_stats])
    efficient_cross_power_support_count = sum([xss.num_eff for xss in x_support_stats])
    coordinate_support_moves = sum([xss.num_xpower_move_coord for xss in x_support_stats])
    cross_power_support_moves = sum([xss.num_xpower_move for xss in x_support_stats])
    print(f"Cross power support: {cross_power_support_count * 100 / support_count} %")
    print(f"Efficient cross power support: {efficient_cross_power_support_count * 100 / cross_power_support_count} %")
    print(f"Cross power support coordination: {coordinate_support_moves * 100 / cross_power_support_moves} %")


def set_orders(orders, game):
    for power, power_orders in orders.items():
        game.set_orders(power, power_orders)


# All code below is adapted from https://github.com/facebookresearch/diplomacy_searchbot/blob/6f72d506677480cf5f91d005c9462dd407c20b69/fairdiplomacy/get_xpower_supports.py
# Licenced under the MIT license

class XPowerSupportStats:

    def __init__(self):
        self.num_orders = 0
        self.num_supports = 0  # total number of supports
        self.num_move = 0  # num support-move
        self.num_xpower = 0  # num xpower supports
        self.num_eff = 0  # num 'effective' xpower supports
        self.num_xpower_move = 0  # num xpower support-move
        self.num_xpower_move_coord = 0  # num coordinated xpower support-move (i.e. the underlying move was taken)


def compute_xpower_supports(phase_orders, game):
    ret = XPowerSupportStats()
    phase = game.get_phase_data()
    state = phase.state
    if not phase.name.endswith("M"):
        return ret
    loc_power = {
        unit.split()[1]: power for power, units in state["units"].items() for unit in units
    }
    # If power owns, e.g., BUL/SC, make it also own BUL. Support targets do
    # not use "/SC" so we need both.
    for loc, power in list(loc_power.items()):
        if "/" in loc:
            loc_land = loc.split("/")[0]
            assert loc_land not in loc_power, (loc_land, loc_power)
            loc_power[loc_land] = power

    all_orders = [o for a in phase_orders.values() for o in a]
    all_order_roots = {order for order in all_orders}
    for power, power_orders in phase_orders.items():
        for order in power_orders:
            ret.num_orders += 1
            order_tokens = order.split()
            is_support = (
                    len(order_tokens) >= 5
                    and order_tokens[2] == "S"
                    and order_tokens[3] in ("A", "F")
            )
            if not is_support:
                continue
            ret.num_supports += 1
            src = order_tokens[4]
            if src not in loc_power:
                raise RuntimeError(f"{order}: {src} not in {loc_power}")
            is_support_move = len(order_tokens) >= 6 and order_tokens[5] == "-"
            if is_support_move:
                ret.num_move += 1

            if loc_power[src] == power:
                continue

            ret.num_xpower += 1

            cf_states = []
            for do_support in (False, True):
                g_cf = deepcopy(game)
                g_cf.build_caches()
                g_cf.set_orders(power, power_orders)

                assert g_cf.get_phase_data().name == phase.name
                if not do_support:
                    hold_order = " ".join(order_tokens[:2] + ["H"])
                    g_cf.set_orders(power, [hold_order])

                g_cf.process()
                assert g_cf.get_phase_data().name != phase.name
                s = g_cf.get_state()
                cf_states.append((s["name"], s["units"], s["retreats"]))

            if cf_states[0] != cf_states[1]:
                ret.num_eff += 1

            if is_support_move:
                ret.num_xpower_move += 1

                underlying_order = " ".join(order_tokens[3:])
                if underlying_order in all_order_roots:
                    ret.num_xpower_move_coord += 1

    return ret
