import ray

from globals import Time
from utils.utils import PickableDefaultDict
import numpy as np


class DefaultTournament:

    def __init__(self, game, episodes=None, callback_list=None):
        self.game = game
        self.episodes = episodes
        self.callback_list = callback_list
        self.agents = self.game.get_players()

    def play_games(self):
        """
        Play multiple rounds of a game. The number of rounds is determined by self.episodes
        :return:
        """
        for episode_nr in range(1, self.episodes + 1):
            self.on_episode_start(episode_nr=episode_nr)
            history = self.play_game(episode_nr=episode_nr)
            self.on_episode_end(history=history, agents=self.agents, episode_nr=episode_nr)

    def play_game(self, episode_nr=1):
        """
        Play a single round of a game
        :param episode_nr:
        :return:
        """
        history = DiplomacyHistory()
        state = self.game.reset()
        step = 1
        while True:
            actions, available_actions = {}, {}
            for agent in self.agents:
                available_actions[agent] = self.game.get_available_actions(agent)
                actions[agent] = agent.next_action(state, available_actions[agent])
            next_state, rewards, done = self.game.step(actions)
            for agent in self.agents:
                history.append_to_history(agent.agent_uid, step, state, available_actions[agent], actions[agent],
                                          rewards[agent],
                                          next_state, done)
            self.on_step_end(agents=self.agents, history=history, episode_nr=episode_nr)
            if done:
                break
            state = next_state
            step += 1
        return history

    def on_episode_start(self, **kwargs):
        if self.callback_list:
            self.callback_list.on_episode_start(**kwargs)

    def on_episode_end(self, **kwargs):
        if self.callback_list:
            self.callback_list.on_episode_end(**kwargs)

    def on_step_end(self, **kwargs):
        if self.callback_list:
            self.callback_list.on_step_end(**kwargs)


class OneStepLearningTournament(DefaultTournament):
    """
    Tournament where the players can learn from a history for every step in the game
    """

    def on_step_end(self, **kwargs):
        for agent in kwargs["agents"]:
            agent.learn(kwargs["history"])
        super().on_step_end(**kwargs)


class BatchLearningTournament(DefaultTournament):
    """
    Tournament where the players can learn from a history after each episode
    """

    def on_episode_end(self, **kwargs):
        for agent in kwargs["agents"]:
            agent.learn(kwargs["history"])
        super().on_episode_end(**kwargs)


class IMPALATournament(DefaultTournament):
    """
    IMPALA Tournament. Controls impala worker processes, receives data from the workers
    and send the data to a learner. Sends updated parameters to the worker processes.
    """

    def __init__(self, game, episodes, callback_list, impala_worker_class, impala_worker_config, impala_workers):
        super().__init__(game, episodes, callback_list)
        RemoteActor = ray.remote(impala_worker_class)
        self.workers = [RemoteActor.remote(i, Time.START_TIME, impala_worker_config) for i in range(impala_workers)]

    def play_games(self):
        # adapted from https://groups.google.com/g/ray-dev/c/dMTLrQAo-1g/m/x-XHNuArAwAJ
        for episode_nr in range(1, self.episodes + 1):
            histories = ray.get([w.play_game.remote() for w in self.workers])
            self.on_episode_end(agents=self.agents, episode_nr=episode_nr, histories=histories)

    def on_episode_end(self, **kwargs):
        for history in kwargs["histories"]:
            for agent in self.agents:
                agent.learn(history)
        self.send_updated_parameters(self.workers)
        super().on_episode_end(**kwargs)

    def send_updated_parameters(self, workers):
        weights = {agent.agent_uid: agent.get_actor_weights() for agent in self.agents}
        for w in workers:
            w.set_weights.remote(weights)


class TournamentFactory:

    @staticmethod
    def create_tournament(tournament_type, game, callback_list, **tournament_config):
        if tournament_type == "ONE_STEP":
            return OneStepLearningTournament(game, tournament_config["episodes"], callback_list)
        elif tournament_type == "BATCH":
            return BatchLearningTournament(game, tournament_config["episodes"], callback_list)
        elif tournament_type == "IMPALA":
            return IMPALATournament(game, tournament_config["episodes"], callback_list,
                                    impala_worker_class=tournament_config["impala_worker_class"],
                                    impala_worker_config=tournament_config["impala_worker_input"],
                                    impala_workers=tournament_config["impala_workers"])
        else:
            return DefaultTournament(game, tournament_config["episodes"], callback_list)


class DiplomacyHistory:

    def __init__(self):
        self.history = PickableDefaultDict([], nested=True)

    def append_to_history(self, agent_uid, step, state, available_actions, action, reward, next_state, done):
        self.history[agent_uid]["step"].append([step])
        self.history[agent_uid]["now_current_state"].append(state[0])
        self.history[agent_uid]["now_prev_state"].append(state[1])
        self.history[agent_uid]["now_prev_state_orders"].append(state[2])
        self.history[agent_uid]["now_season"].append(state[3])
        self.history[agent_uid]["now_build_numbers"].append(state[4])
        self.history[agent_uid]["available_actions"].append(available_actions[0])
        self.history[agent_uid]["power"].append(available_actions[1])
        self.history[agent_uid]["action"].append(action)
        self.history[agent_uid]["reward"].append([reward])
        self.history[agent_uid]["next_current_state"].append(next_state[0])
        self.history[agent_uid]["next_prev_state"].append(next_state[1])
        self.history[agent_uid]["next_prev_state_orders"].append(next_state[2])
        self.history[agent_uid]["next_season"].append(next_state[3])
        self.history[agent_uid]["next_build_numbers"].append(next_state[4])
        self.history[agent_uid]["done"].append([done])

    def get_current_step(self, agent_uid):
        return np.array([self.history[agent_uid]["step"][-1]])

    def get_current_state(self, agent_uid):
        return [self.history[agent_uid]["now_current_state"][-1],
                self.history[agent_uid]["now_prev_state"][-1],
                self.history[agent_uid]["now_prev_state_orders"][-1],
                self.history[agent_uid]["now_season"][-1],
                self.history[agent_uid]["now_build_numbers"][-1]]

    def get_currently_available_actions(self, agent_uid):
        return [self.history[agent_uid]["available_actions"][-1],
                self.history[agent_uid]["power"][-1]]

    def get_current_action(self, agent_uid):
        return np.array([self.history[agent_uid]["action"][-1]])

    def get_next_state(self, agent_uid):
        return [self.history[agent_uid]["next_current_state"][-1],
                self.history[agent_uid]["next_prev_state"][-1],
                self.history[agent_uid]["next_prev_state_orders"][-1],
                self.history[agent_uid]["next_season"][-1],
                self.history[agent_uid]["next_build_numbers"][-1]]

    def get_current_reward(self, agent_uid):
        return np.array([self.history[agent_uid]["reward"][-1]])

    def is_currently_done(self, agent_uid):
        return np.array([self.history[agent_uid]["done"][-1]])

    def get_steps(self, agent_uid):
        return np.array(self.history[agent_uid]["step"])

    def get_next_states(self, agent_uid):
        return [np.concatenate(self.history[agent_uid]["next_current_state"]),
                np.concatenate(self.history[agent_uid]["next_prev_state"]),
                np.concatenate(self.history[agent_uid]["next_prev_state_orders"]),
                np.concatenate(self.history[agent_uid]["next_season"]),
                np.concatenate(self.history[agent_uid]["next_build_numbers"])]

    def get_states(self, agent_uid):
        return [np.concatenate(self.history[agent_uid]["now_current_state"]),
                np.concatenate(self.history[agent_uid]["now_prev_state"]),
                np.concatenate(self.history[agent_uid]["now_prev_state_orders"]),
                np.concatenate(self.history[agent_uid]["now_season"]),
                np.concatenate(self.history[agent_uid]["now_build_numbers"])]

    def get_actions(self, agent_uid):
        return np.array(self.history[agent_uid]["action"])

    def get_rewards(self, agent_uid):
        return np.array(self.history[agent_uid]["reward"])

    def get_available_actions(self, agent_uid):
        return [np.concatenate(self.history[agent_uid]["available_actions"]),
                np.concatenate(self.history[agent_uid]["power"])]

    def is_done(self, agent_uid):
        return np.array(self.history[agent_uid]["done"])
