import unittest
from unittest.mock import Mock

from tournament.rl_tournament import DiplomacyHistory, DefaultTournament, OneStepLearningTournament
import numpy as np


class TestDefaultTournament(unittest.TestCase):

    def setUp(self):
        mock_game = Mock()
        self.player1 = Mock()
        self.player1.next_action = Mock(return_value=[1])
        player2 = Mock()
        player2.next_action = Mock(return_value=[2])
        mock_game.get_players = Mock(return_value=[self.player1, player2])
        self.state = [np.random.uniform(size=[1, 7, 31]), np.random.uniform(size=[1, 7, 31]),
                      np.random.uniform(size=[1, 7]), np.random.uniform(size=[1, 7]),
                      np.random.uniform(size=[1, 7])]
        mock_game.reset = Mock(return_value=self.state)
        mock_game.step = Mock(return_value=(self.state, {self.player1: 2, player2: 3}, True))
        mock_game.get_available_actions = Mock(return_value=[[1], [2]])
        self.mock_callback_list = Mock()
        self.mock_callback_list.on_step_end = Mock()
        self.mock_callback_list.on_episode_end = Mock()
        self.mock_callback_list.on_episode_start = Mock()
        self.tournament = DefaultTournament(mock_game, callback_list=self.mock_callback_list)

    def test_play_game(self):
        history = self.tournament.play_game()
        self.mock_callback_list.on_episode_end.assert_not_called()
        self.mock_callback_list.on_episode_start.assert_not_called()
        self.mock_callback_list.on_step_end.assert_called()
        np.testing.assert_equal(self.state, history.get_states(self.player1.agent_uid))
        np.testing.assert_equal(self.state, history.get_states(self.player1.agent_uid))

    def test_on(self):
        self.tournament.on_episode_end()
        self.mock_callback_list.on_episode_end.assert_called()
        self.tournament.on_episode_start()
        self.mock_callback_list.on_episode_start.assert_called()
        self.tournament.on_step_end()
        self.mock_callback_list.on_step_end.assert_called()

    def test_on_no_callback_list(self):
        self.tournament.callback_list = None
        self.tournament.on_episode_end()
        self.tournament.on_step_end()
        self.tournament.on_episode_start()


class TestOneStepLearningTournament(unittest.TestCase):

    def setUp(self):
        mock_game = Mock()
        self.player1 = Mock()
        self.player1.next_action = Mock(return_value=[1])
        self.player1.learn = Mock()
        self.player2 = Mock()
        self.player2.next_action = Mock(return_value=[2])
        self.player2.learn = Mock()
        mock_game.get_players = Mock(return_value=[self.player1, self.player2])
        self.state = [np.random.uniform(size=[1, 7, 31]), np.random.uniform(size=[1, 7, 31]),
                      np.random.uniform(size=[1, 7]), np.random.uniform(size=[1, 7]),
                      np.random.uniform(size=[1, 7])]
        mock_game.reset = Mock(return_value=self.state)
        mock_game.step = Mock(return_value=(self.state, {self.player1: 2, self.player2: 3}, True))
        mock_game.get_available_actions = Mock(return_value=[[1], [2]])
        self.mock_callback_list = Mock()
        self.mock_callback_list.on_step_end = Mock()
        self.mock_callback_list.on_episode_end = Mock()
        self.mock_callback_list.on_episode_start = Mock()
        self.tournament = OneStepLearningTournament(mock_game, episodes=3, callback_list=self.mock_callback_list)

    def test_play_game(self):
        self.tournament.play_game()
        self.player1.next_action.assert_called()
        self.player2.next_action.assert_called()
        self.mock_callback_list.on_step_end.assert_called()
        self.mock_callback_list.on_episode_end.assert_not_called()
        self.mock_callback_list.on_episode_start.assert_not_called()

    def test_play_games(self):
        self.tournament.play_games()
        self.player1.next_action.assert_called()
        self.player2.next_action.assert_called()
        self.mock_callback_list.on_step_end.assert_called()
        self.mock_callback_list.on_episode_end.assert_called()
        self.mock_callback_list.on_episode_start.assert_called()


class TestDiplomacyHistory(unittest.TestCase):

    def test_current_information(self):
        history = DiplomacyHistory()
        player1 = "Player1"
        for i in range(100):
            state = np.array([[i], [i], [i], [i], [i]])
            available_actions = np.array([[i], [i]])
            next_state = state + 1
            history.append_to_history(player1, i, state, available_actions, i, i, next_state, i == 99)
            np.testing.assert_equal(state, history.get_current_state(player1))
            self.assertEqual(i, history.get_current_step(player1))
            np.testing.assert_equal(next_state, history.get_next_state(player1))
            self.assertEqual(i, history.get_current_reward(player1))
            self.assertEqual(i, history.get_current_action(player1)[0])
            self.assertEqual(history.is_currently_done(player1), i == 99)


if __name__ == '__main__':
    unittest.main()
