from absl import flags, app
import os

from agents import AgentFactory
from callbacks.callbacks import CallbackFactory
from games import GameFactory
from neural_networks.networks import NeuralNetworkFactory
from tournament.rl_tournament import TournamentFactory
from utils.utils import get_config
from globals import Time

os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "true"
os.environ["TF_GPU_ALLOCATOR"] = "cuda_malloc_async"

FLAGS = flags.FLAGS
flags.DEFINE_string("game_config_path", None,
                    "a path to a file which specifies properties of the game and the agents who will play a game of Diplomacy")

ALGORITHM_NAME_TO_AGENT_TYPE = {"IMPALA_A2C": "BATCH_A2C", "IMPALA_LOLA": "BATCH_LOLA",
                                "A2C": "A2C", "LOLA": "LOLA", "BATCH_A2C": "BATCH_A2C", "BATCH_LOLA":
                                    "BATCH_LOLA"}
ALGORITHM_NAME_TO_TOURNAMENT_TYPE = {"IMPALA_A2C": "IMPALA", "IMPALA_LOLA": "IMPALA",
                                     "A2C": "ONE_STEP", "LOLA": "ONE_STEP", "BATCH_A2C": "BATCH", "BATCH_LOLA":
                                         "BATCH"}


def main(argv):
    Time.set_start_time()
    config = get_config(FLAGS.game_config_path)
    tournament_type = ALGORITHM_NAME_TO_TOURNAMENT_TYPE[config["algorithm"]["name"]]
    diplomacy_learner = setup_training(config, tournament_type)
    diplomacy_learner.play_games()


def setup_training(config, tournament_type):
    players = create_list_of_agents(config)
    callback_list = CallbackFactory.create_game_callbacks(config["game"]["name"],
                                                          file_name="train_log.csv",
                                                          **config)
    game = GameFactory.create_game(players, callback_list, **config["game"])
    rl_callbacks = CallbackFactory.create_tournament_callbacks(config["game"]["name"],
                                                               test_tournament_configs=config["test_tournaments"],
                                                               **config)
    return TournamentFactory.create_tournament(tournament_type, game, rl_callbacks,
                                               impala_worker_class=IMPALAWorker,
                                               impala_worker_input=config,
                                               **config["algorithm"])


def create_list_of_agents(config):
    policy_net, value_net = NeuralNetworkFactory.create_neural_networks(config["game"]["name"],
                                                                        config["neural_network_parameters"],
                                                                        map_name=config["game"]["map_name"])
    agent_type = ALGORITHM_NAME_TO_AGENT_TYPE[config["algorithm"]["name"]]
    agents = AgentFactory.create_list_of_agents(agent_type, pnet=policy_net, vnet=value_net,
                                                agent_configs=config["agents"])
    return agents


class IMPALAWorker(object):

    def __init__(self, uid, start_time, config):
        super().__init__()
        self.uid = uid
        config["sub_dir"] = "" + start_time + f"/impala_worker-{self.uid}"
        self.tournament = setup_training(config, "DEFAULT")
        self.agents = self.tournament.agents
        self.episode_nr = 1

    def play_game(self):
        print(f"Worker {self.uid}, episode {self.episode_nr}")
        history = self.tournament.play_game()
        self.episode_nr += 1
        return history

    def set_weights(self, weights):
        for agent in self.agents:
            agent.set_actor_weights(weights[agent.agent_uid])


if __name__ == "__main__":
    app.run(main)
